# -*- coding: utf-8 -*-
"""
cross platform utilities
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import os
import pipes
import six
import sys
import platform
import subprocess
import shlex
from os.path import exists, normpath, basename, dirname, join, expanduser
from utool import util_inject
from utool._internal import meta_util_cplat
from utool._internal.meta_util_path import unixpath, truepath
from six.moves import zip
print, rrr, profile = util_inject.inject2(__name__)
print_ = util_inject.make_module_write_func(__name__)

try:
    import pathlib
    HAVE_PATHLIB = True
except ImportError:
    HAVE_PATHLIB = False

COMPUTER_NAME = platform.node()

OS_TYPE = meta_util_cplat.OS_TYPE
WIN32  = meta_util_cplat.WIN32
LINUX  = meta_util_cplat.LINUX
DARWIN = meta_util_cplat.DARWIN
UNIX = not WIN32


# class Win32Err(object):
#     """
#     References:
#         https://msdn.microsoft.com/en-us/library/cc231199.aspx
#         https://docs.python.org/3/library/os.html#os.EX_OK
#     """
#     ERROR_SUCCESS = 0x0
#     ERROR_INVALID_FUNCTION = 0x1
#     ERROR_FATAL_APP_EXIT = 0x2C9

# Define standarized exit codes
EXIT_SUCCESS = 0
EXIT_FAILURE = 1

LIB_EXT_LIST = ['.so', '.dll', '.dylib', '.pyd']

LIB_DICT = {
    'win32': '.dll',
    'linux': '.so',
    'darwin': '.dylib',
}


PYLIB_DICT = {
    'win32': '.pyd',
    'linux': '.so',
    'darwin': '.dylib',
}


def get_plat_specifier():
    """
    Standard platform specifier used by distutils
    """
    import setuptools  # NOQA
    import distutils
    plat_name = distutils.util.get_platform()
    plat_specifier = ".%s-%s" % (plat_name, sys.version[0:3])
    if hasattr(sys, 'gettotalrefcount'):
        plat_specifier += '-pydebug'
    return plat_specifier


def in_pyinstaller_package():
    """
    References:
        http://stackoverflow.com/questions/22472124/what-is-sys-meipass-in-python
        http://stackoverflow.com/questions/7674790/bundling-data-files-with-pyinstaller-onefile
    """
    return hasattr(sys, '_MEIPASS')


def get_system_python_library():
    """
    FIXME; hacky way of finding python library. Not cross platform yet.
    """
    import os
    import utool as ut
    from os.path import basename, realpath
    pyname = basename(realpath(sys.executable))
    ld_library_path = os.environ['LD_LIBRARY_PATH']
    libdirs = [x for x in ld_library_path.split(os.pathsep) if x] + ['/usr/lib']
    libfiles = ut.flatten([ut.glob(d, '*' + ut.get_lib_ext(), recursive=True) for d in libdirs])
    python_libs = [realpath(f) for f in libfiles if 'lib' + pyname in basename(f)]
    python_libs = ut.unique_ordered(python_libs)
    assert len(python_libs) == 1, str(python_libs)
    return python_libs[0]


def get_free_diskbytes(dir_):
    r"""
    Args:
        dir_ (str):

    Returns:
        int: bytes_ folder/drive free space (in bytes)

    References::
        http://stackoverflow.com/questions/51658/cross-platform-space-remaining-on-volume-using-python
        http://linux.die.net/man/2/statvfs

    CommandLine:
        python -m utool.util_cplat --exec-get_free_diskbytes
        python -m utool.util_cplat --exec-get_free_diskbytes --dir /media/raid
        python -m utool.util_cplat --exec-get_free_diskbytes --dir E:

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_cplat import *  # NOQA
        >>> import utool as ut
        >>> dir_ = ut.get_argval('--dir', type_=str, default=ut.truepath('~'))
        >>> bytes_ = get_free_diskbytes(dir_)
        >>> result = ('bytes_ = %s' % (str(bytes_),))
        >>> print(result)
        >>> print('Unused space in %r = %r' % (dir_, ut.byte_str2(bytes_)))
        >>> print('Total space in %r = %r' % (dir_, ut.byte_str2(get_total_diskbytes(dir_))))
    """
    if WIN32:
        import ctypes
        free_bytes = ctypes.c_ulonglong(0)
        outvar = ctypes.pointer(free_bytes)
        dir_ptr = ctypes.c_wchar_p(dir_)
        ctypes.windll.kernel32.GetDiskFreeSpaceExW(dir_ptr, None, None, outvar)
        bytes_ = free_bytes.value
        return bytes_
    else:
        st = os.statvfs(dir_)
        # blocks avaiable * block size
        bytes_ = st.f_bavail * st.f_frsize
        #bytes_ = st.f_bfree * st.f_frsize  # includes root only space
        return bytes_


def get_total_diskbytes(dir_):
    if WIN32:
        import ctypes
        total_bytes = ctypes.c_ulonglong(0)
        outvar = ctypes.pointer(total_bytes)
        dir_ptr = ctypes.c_wchar_p(dir_)
        ctypes.windll.kernel32.GetDiskFreeSpaceExW(dir_ptr, None, outvar, None)
        bytes_ = total_bytes.value
        return bytes_
    else:
        #raise NotImplementedError('')
        st = os.statvfs(dir_)
        # blocks total * block size
        bytes_ = st.f_blocks * st.f_frsize
        return bytes_


def chmod_add_executable(fpath, group=True, user=True):
    """
    References:
        http://stackoverflow.com/questions/15607903/python-module-os-chmodfile-664-does-not-change-the-permission-to-rw-rw-r-bu
        http://www.tutorialspoint.com/python/os_chmod.htm
        https://en.wikipedia.org/wiki/Chmod
    """
    import stat
    orig_mode = os.stat(fpath).st_mode
    new_mode = orig_mode
    if group:
        new_mode |= stat.S_IXGRP
    if user:
        # new_mode |= stat.S_IXUSR | stat.S_IEXEC
        new_mode |= stat.S_IXGRP | stat.S_IEXEC
    os.chmod(fpath, new_mode)


def chmod(fpath, option):
    import stat
    orig_mode = os.stat(fpath).st_mode
    new_mode = orig_mode
    if option == '+x':
        # Hack
        new_mode |= stat.S_IEXEC
    if option == 'g+x':
        new_mode |= stat.S_IXGRP

    os.chmod(fpath, new_mode)


def is_file_writable(fpath):
    return os.access(fpath, os.W_OK)


def is_file_executable(fpath):
    return os.access(fpath, os.X_OK)


def get_file_info(fpath, with_fpath=False, nice=True):
    from utool import util_time
    import os
    import time
    from collections import OrderedDict
    statbuf = os.stat(fpath)

    from pwd import getpwuid
    owner = getpwuid(os.stat(fpath).st_uid).pw_name

    info = OrderedDict([
        ('filesize', get_file_nBytes_str(fpath)),
        ('last_modified', statbuf.st_mtime),
        ('last_accessed', statbuf.st_atime),
        ('created', statbuf.st_ctime),
        ('owner', owner)
    ])
    if nice:
        for k in ['last_modified', 'last_accessed', 'created']:
            info[k] = util_time.unixtime_to_datetimestr(info[k], isutc=False)
            info[k] += (' ' + time.tzname[0])

    if with_fpath:
        info['fpath'] = fpath
    return info
    #print "Modification time:",statbuf.st_mtime


def get_file_nBytes(fpath):
    return os.path.getsize(fpath)


def get_file_nBytes_str(fpath):
    from utool import util_str
    return util_str.byte_str2(os.path.getsize(fpath))


def get_disk_space(start_path='.'):
    """
    References:
        http://stackoverflow.com/questions/1392413/calculating-a-directory-size-using-python
    """
    total_size = 0
    for root, dname_list, fname_list in os.walk(start_path):
        for fname in fname_list:
            fpath = os.path.join(root, fname)
            try:
                total_size += os.path.getsize(fpath)
            except OSError:
                pass
    return total_size


def get_dir_diskspaces(dir_):
    from utool import util_path
    path_list = util_path.ls(dir_)
    nBytes_list = [get_disk_space(path) for path in path_list]
    spacetup_list = sorted(list(zip(nBytes_list, path_list)))
    return spacetup_list


def print_dir_diskspace(dir_):
    import utool
    spacetup_list = sorted(get_dir_diskspaces(dir_))
    nBytes_list = [tup[0] for tup in spacetup_list]
    path_list   = [tup[1] for tup in spacetup_list]
    space_list = map(utool.byte_str2, nBytes_list)
    n = max(map(len, space_list))
    fmtstr = ('%' + str(n) + 's')
    space_list2 = [fmtstr % space for space in space_list]
    tupstr_list = ['%s %s' % (space2, path) for space2, path in zip(space_list2, path_list)]
    print('\n'.join(tupstr_list))


def get_lib_ext():
    return LIB_DICT[OS_TYPE]


def get_pylib_ext():
    return PYLIB_DICT[OS_TYPE]


def python_executable(check=True, short=False):
    r"""
    Args:
        short (bool): (default = False)

    Returns:
        str:

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_cplat import *  # NOQA
        >>> short = False
        >>> result = python_executable(short)
        >>> print(result)
    """
    if not check:
        python_exe = 'python'
    else:
        from os.path import isdir
        python_exe_long = unixpath(sys.executable)
        python_exe = python_exe_long
        if short:
            python_exe_short = basename(python_exe_long)
            found = search_env_paths(python_exe_short, key_list=['PATH'],
                                     verbose=False)
            found = [f for f in found if not isdir(f)]
            if len(found) > 0:
                if found[0] == python_exe_long:
                    # Safe to use the short name in this env
                    python_exe = python_exe_short
    return python_exe


def ls_libs(dpath):
    from utool import util_list
    from utool import util_path
    lib_patterns = get_dynamic_lib_globstrs()
    libpaths_list = [util_path.ls(dpath, pat) for pat in lib_patterns]
    libpath_list = util_list.flatten(libpaths_list)
    return libpath_list


def get_dynlib_dependencies(lib_path):
    """
    Executes tools for inspecting dynamic library dependencies depending on the
    current platform.
    """
    if LINUX:
        ldd_fpath = '/usr/bin/ldd'
        depend_out, depend_err, ret = cmd(ldd_fpath, lib_path, verbose=False)
    elif DARWIN:
        otool_fpath = '/opt/local/bin/otool'
        depend_out, depend_err, ret = cmd(otool_fpath, '-L', lib_path, verbose=False)
    elif WIN32:
        depend_out, depend_err, ret = cmd('objdump', '-p', lib_path, verbose=False)
        #fnmatch.filter(depend_out.split('\n'), '*DLL*')
        relevant_lines = [line for line in depend_out.splitlines() if 'DLL Name:' in line]
        depend_out = '\n'.join(relevant_lines)
    assert ret == 0, 'bad dependency check'
    return depend_out
    # objdump -p C:\Python27\Lib\site-packages\PIL\_imaging.pyd | grep dll
    # dumpbin /dependents C:\Python27\Lib\site-packages\PIL\_imaging.pyd
    # depends /c /a:1 /f:1 C:\Python27\Lib\site-packages\PIL\_imaging.pyd


def get_dynlib_exports(lib_path):
    """
    Executes tools for inspecting dynamic library dependencies depending on the
    current platform. Returns the names of callable functions.

    Args:
        lib_path (str):

    Returns:
        str: depend_out

    CommandLine:
        python -m utool.util_cplat --test-get_dynlib_exports

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_cplat import *  # NOQA
        >>> lib_path = '/home/joncrall/venv/local/lib/python2.7/site-packages/pyflann/lib/libflann.so'
        >>> depend_out = get_dynlib_exports(lib_path)
        >>> result = ('depend_out = %s' % (str(depend_out),))
        >>> print(result)
    """
    if LINUX:
        '''
        nm_fpath = '/usr/bin/nm'
        exportssout, err, ret = cmd(nm_fpath, '-D', lib_path, '|', 'c++filt', verbose=False)
        lines = exportssout.split('\n')
        #lines = [line[19:] for line in line]
        others = []
        info = []
        for line in lines:
            if line == '':
                continue
            line = ut.remove_doublspaces(line)
            words = line.split(' ')
            if len(words) > 2:
                # address, type_, rest
                rest = ' '.join(words[2:])
                info.append((rest, words[0], words[1]))
            else:
                others.append(line)

        # remove duplicate address spaces
        info = ut.unique_ordered(info)
        # remove stdlib
        info = [line for line in info if 'std::' not in line[0]]
        info = [line for line in info if not line[0].startswith('typeinfo')]
        info = [line for line in info if not line[0].startswith('vtable')]
        info = [line for line in info if 'flann' in line[0]]
        info = [line for line in info if 'flann_' in line[0]]

        info2 = []
        for rest, loc, type_ in info:
            parts = rest.split(' ')
            rettype = parts[0]
            rest2 = ' '.join(parts[1:])
            if not rest2.startswith('__'):
                info2.append((rettype, rest2, type_))
                #info2.append((rettype, rest2, type_, loc))

        len([line for line in info if 'flann' in line[0]])

        len([(line.split(' ')[0], line.split(' ')[1], ' '.join(line.split(' ')[2:])) for line in lines])
        len([line for line in lines if line.startswith('flann::')])
        len([line for line in lines if 'flann_' in line])
        len([line for line in lines if not line.endswith(')') and 'flann_' in line])
        # HACK: FIND A CORRECT PARSING
        return info2
        '''
    elif DARWIN:
        otool_fpath = '/opt/local/bin/otool'
        exportssout, err, ret = cmd(otool_fpath, '-L', lib_path, verbose=False)
        #TODO
    elif WIN32:
        exportssout, err, ret = cmd('objdump', '-p', lib_path, verbose=False)
        #TODO
        #fnmatch.filter(depend_out.split('\n'), '*DLL*')
        #relevant_lines = [line for line in depend_out.splitlines() if 'DLL Name:' in line]
        #depend_out = '\n'.join(relevant_lines)
    assert ret == 0, 'bad dependency check'
    return exportssout


def get_dynamic_lib_globstrs():
    return ['*' + libext for libext in LIB_EXT_LIST]


def get_computer_name():
    """ Returns machine name """
    return COMPUTER_NAME


def get_user_name():
    """ Returns user homefolder name """
    return basename(truepath('~'))


def get_install_dirs():
    if WIN32:
        return [r'C:\Program Files', r'C:\Program Files (x86)']
    else:
        return ['/usr/bin', '/usr/local/bin', '~/bin']


def getroot():
    root = {
        'win32': 'C:\\',  # HACK
        'linux': '/',
        'darwin': '/',
    }[OS_TYPE]
    return root


def startfile(fpath, detatch=True, quote=False, verbose=False, quiet=True):
    """ Uses default program defined by the system to open a file.

    References:
        http://stackoverflow.com/questions/2692873/quote-posix-shell-special-characters-in-python-output

    """
    print('[cplat] startfile(%r)' % fpath)
    fpath = normpath(fpath)
    # print('[cplat] fpath=%s' % fpath)
    if not exists(fpath):
        raise Exception('Cannot start nonexistant file: %r' % fpath)
    #if quote:
    #    fpath = '"%s"' % (fpath,)
    if not WIN32:
        fpath = pipes.quote(fpath)
    if LINUX:
        #out, err, ret = cmd(['xdg-open', fpath], detatch=True)
        outtup = cmd(('xdg-open', fpath), detatch=detatch, verbose=verbose, quiet=quiet)
        #outtup = cmd('xdg-open', fpath, detatch=detatch)
    elif DARWIN:
        outtup = cmd(('open', fpath), detatch=detatch, verbose=verbose, quiet=quiet)
    elif WIN32:
        os.startfile(fpath)
    else:
        raise RuntimeError('Unknown Platform')
    if outtup is not None:
        out, err, ret = outtup
        if not ret:
            raise Exception(out + ' -- ' + err)
    pass


def geteditor():
    return 'gvim'  # HACK (use util_profile)


def editfile(fpath):
    """ Runs gvim. Can also accept a module / class / function """
    if not isinstance(fpath, six.string_types):
        from six import types
        print('Rectify to module fpath = %r' % (fpath,))
        if isinstance(fpath, types.ModuleType):
            fpath = fpath.__file__
        else:
            fpath =  sys.modules[fpath.__module__].__file__
        fpath_py = fpath.replace('.pyc', '.py')
        if exists(fpath_py):
            fpath = fpath_py

    print('[cplat] startfile(%r)' % fpath)
    if not exists(fpath):
        raise Exception('Cannot start nonexistant file: %r' % fpath)
    if LINUX:
        out, err, ret = cmd(geteditor(), fpath, detatch=True)
        if not ret:
            raise Exception(out + ' -- ' + err)
    elif DARWIN:
        out, err, ret = cmd(geteditor(), fpath, detatch=True)
        if not ret:
            raise Exception(out + ' -- ' + err)
    else:
        out, err, ret = cmd(geteditor(), fpath, detatch=True)
        if not ret:
            raise Exception(out + ' -- ' + err)
        #os.startfile(fpath)
    pass


def view_file_in_directory(fpaths):
    import utool as ut
    fpaths = ut.ensure_iterable(fpaths)
    fnames = [basename(f) for f in fpaths]
    dpaths = [dirname(f) for f in fpaths]
    dpath_to_fnames = ut.group_items(fnames, dpaths)
    for dpath, fnames in dpath_to_fnames.items():
        ut.view_directory(dpath, fnames[0], verbose=False)


def view_directory(dname=None, fname=None, verbose=True):
    """
    View a directory in the operating system file browser. Currently supports
    windows explorer, mac open, and linux nautlius.

    Args:
        dname (str): directory name
        fname (str): a filename to select in the directory (nautlius only)
        verbose (bool):

    CommandLine:
        python -m utool.util_cplat --test-view_directory

    Example:
        >>> # DISABLE_DOCTEST
        >>> # DOCTEST_DISABLE
        >>> from utool.util_cplat import *  # NOQA
        >>> import utool as ut
        >>> dname = ut.truepath('~')
        >>> verbose = True
        >>> view_directory(dname, verbose)

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_cplat import *  # NOQA
        >>> import utool as ut
        >>> base = ut.ensure_app_cache_dir('utool', 'test_vd')
        >>> dirs = [
        >>>     '',
        >>>     'dir1',
        >>>     'has space',
        >>>     'space at end ',
        >>>     ' space at start ',
        >>>     '"quotes and spaces"',
        >>>     "'single quotes and spaces'",
        >>>     'Frogram Piles (y2K)',
        >>> ]
        >>> dirs_ = [ut.ensuredir(join(base, d)) for d in dirs]
        >>> for dname in dirs_:
        >>>     ut.view_directory(dname, verbose=False)
        >>> fpath = join(base, 'afile.txt')
        >>> ut.touch(fpath)
        >>> ut.view_directory(base, fpath, verbose=False)
    """
    from utool.util_arg import STRICT
    from utool.util_path import checkpath
    # from utool.util_str import SINGLE_QUOTE, DOUBLE_QUOTE

    if HAVE_PATHLIB and isinstance(dname, pathlib.Path):
        dname = str(dname)

    if verbose:
        print('[cplat] view_directory(%r) ' % dname)
    dname = os.getcwd() if dname is None else dname
    open_prog = {
        'win32': 'explorer.exe',
        'linux': 'nautilus',
        'darwin': 'open'
    }[OS_TYPE]
    dname = normpath(dname)
    if STRICT:
        assert checkpath(dname, verbose=verbose), 'directory doesnt exit'
    if fname is not None and OS_TYPE == 'linux':
        arg = join(dname, fname)
    else:
        arg = dname
    # if ' ' in dname and not dname.startswith((SINGLE_QUOTE, DOUBLE_QUOTE)):
    #     # Ensure quotations
    #     dname = '"%s"' % dname
    # if not WIN32:
    #     arg = dname
    #     # arg = subprocess.list2cmdline([dname])
    #     # arg = pipes.quote(dname)
    # else:
    #     arg = dname
    # spawn and detatch process
    args = (open_prog, arg)
    print(subprocess.list2cmdline(args))
    subprocess.Popen(args)
    # print('[cplat] exit view directory')

# Alias
vd = view_directory

get_resource_dir = meta_util_cplat.get_resource_dir

get_app_resource_dir = meta_util_cplat.get_app_resource_dir


def platform_cache_dir():
    """
    Returns a directory which should be writable for any application
    This should be used for temporary deletable data.
    """
    if WIN32:  # nocover
        dpath_ = '~/AppData/Local'
    elif LINUX:  # nocover
        dpath_ = '~/.cache'
    elif DARWIN:  # nocover
        dpath_  = '~/Library/Caches'
    else:  # nocover
        raise NotImplementedError('Unknown Platform  %r' % (sys.platform,))
    dpath = normpath(expanduser(dpath_))
    return dpath


def get_app_cache_dir(appname, *args):
    r"""
    Returns a writable directory for an application.
    This should be used for temporary deletable data.

    Args:
        appname (str): the name of the application
        *args: any other subdirectories may be specified

    Returns:
        str: dpath: writable cache directory
    """
    dpath = join(platform_cache_dir(), appname, *args)
    return dpath


def ensure_app_cache_dir(appname, *args):
    """
    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_cplat import *  # NOQA
        >>> import utool as ut
        >>> dpath = ut.ensure_app_cache_dir('utool')
        >>> assert exists(dpath)
    """
    import utool as ut
    dpath = get_app_cache_dir(appname, *args)
    ut.ensuredir(dpath)
    return dpath


def ensure_app_resource_dir(*args, **kwargs):
    import utool as ut
    app_resource_dir = get_app_resource_dir(*args, **kwargs)
    ut.ensuredir(app_resource_dir)
    return app_resource_dir


def shell(*args, **kwargs):
    """
    Dangerous. Take out of production code
    """
    kwargs['shell'] = True
    return cmd(*args, **kwargs)


def __parse_cmd_kwargs(kwargs):
    verbose = kwargs.get('verbose', True)
    detatch = kwargs.get('detatch', False)
    #shell   = kwargs.get('shell', False)
    # seems like linux needs the shell to work well
    # maybe thats because I'm a windows admin
    # FIXME: Turn shell off by default and fix __parse_cmd_args
    shell   = kwargs.get('shell', LINUX or DARWIN)
    # TODO: gksudo
    sudo    = kwargs.get('sudo', False)
    # pads stdout of cmd before and after
    # TODO: rename separate to something else
    silence    = kwargs.get('silence', False)
    quiet    = kwargs.get('quiet', False)
    pad_stdout    = kwargs.get('pad_stdout', not (silence or quiet))
    return verbose, detatch, shell, sudo, pad_stdout


def __parse_cmd_args(args, sudo, shell):
    """
    When shell is True, Popen will only accept strings. No tuples
    Shell really should not be true.

    Returns:
        args suitable for subprocess.Popen

        I'm not quite sure what those are yet. Plain old string seem to work
        well? But I remember needing shlex at some point.

    CommandLine:
        python -m utool.util_cplat --test-__parse_cmd_args

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_cplat import *  # NOQA
        >>> # build test data
        >>> args = 'echo "hello world"'
        >>> sudo = False
        >>> shell = False
        >>> # execute function
        >>> args = __parse_cmd_args(args, sudo, shell)
        >>> # verify results
        >>> result = str(args)
        >>> print(result)
    """
    # Case where tuple is passed in as only argument
    if isinstance(args, tuple) and len(args) == 1 and isinstance(args[0], tuple):
        args = args[0]

    if shell:
        # When shell is True, ensure args is a string
        if isinstance(args, six.string_types):
            pass
        elif  isinstance(args, (list, tuple)) and len(args) > 1:
            args = ' '.join(args)
        elif isinstance(args, (list, tuple)) and len(args) == 1:
            if isinstance(args[0], (tuple, list)):
                args = ' '.join(args)
            elif isinstance(args[0], six.string_types):
                args = args[0]
    else:
        # When shell is False, ensure args is a tuple
        if isinstance(args, six.string_types):
            args = shlex.split(args, posix=not WIN32)
        elif isinstance(args, (list, tuple)):
            if len(args) > 1:
                args = tuple(args)
            elif len(args) == 1:
                if isinstance(args[0], (tuple, list)):
                    args = tuple(args[0])
                elif isinstance(args[0], six.string_types):
                    args = shlex.split(args[0], posix=not WIN32)
    if sudo is True:
        if not WIN32:
            if shell:
                args = 'sudo ' + args
            else:
                args = tuple(['sudo']) + tuple(args)
            #if isinstance(args, six.string_types):
            #    args = shlex.split(args)
            #args = ['sudo'] + args
            ## using sudo means we need to use a single string I believe
            #args = ' '.join(args)
        else:
            # TODO: strip out sudos
            pass
    # HACK FOR WINDOWS AGAIN
    # makes  this command work:
    # python -c "import utool as ut; ut.cmd('build\\hesaffexe.exe ' + ut.grab_test_imgpath('star.png'))"
    # and this should still work
    # python -c "import utool as ut; ut.cmd('build\\hesaffexe.exe', ut.grab_test_imgpath('star.png'))"
    if WIN32:
        if len(args) == 1 and isinstance(args[0], six.string_types):
            args = shlex.split(args[0], posix=not WIN32)
    return args


def run_realtime_process(exe, shell=False):
    proc = subprocess.Popen(exe, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=shell)
    while(True):
        # WARNING: this can cause deadlocks apparently if the OS pipe buffers
        # fill up.
        retcode = proc.poll()  # returns None while subprocess is running
        line = proc.stdout.readline()
        yield line
        if retcode is not None:
            return
            # raise StopIteration('process finished')


def _run_process(proc):
    while True:
        # returns None while subprocess is running
        retcode = proc.poll()
        line = proc.stdout.readline()
        yield line
        if retcode is not None:
            # The program has a return code, so its done executing.
            # Grab any remaining data in stdout
            for line in proc.stdout.readlines():
                yield line
            return
            # raise StopIteration('process finished')


def quote_single_command(cmdstr):
    if ' ' in cmdstr:
        return '\'' + cmdstr + '\''
    return cmdstr


def cmd(*args, **kwargs):
    r""" A really roundabout way to issue a system call

    # FIXME: This function needs some work
    # It should work without a hitch on windows or unix.
    # It should be able to spit out stdout in realtime.
    # Should be able to configure detatchment, shell, and sudo.

    FIXME:
        on a mac ut.cmd('/Users/joncrall/Library/Application Support/ibeis/tomcat/bin/shutdown.sh') will fail due to spaces


    Kwargs:
        quiet (bool) :
        silence (bool) :
        verbose (bool) :
        detatch (bool) :
        shell (bool) :
        sudo (bool) :
        pad_stdout (bool) :
        dryrun (bool) :

    Returns:
        tuple: (None, None, None)

    CommandLine:
        python -m utool.util_cplat --test-cmd
        python -m utool.util_cplat --test-cmd:0
        python -m utool.util_cplat --test-cmd:1
        python -m utool.util_cplat --test-cmd:2
        python -m utool.util_cplat --test-cmd:1 --test-sudo
        python -m utool.util_cplat --test-cmd:2 --test-sudo

    Example0:
        >>> # ENABLE_DOCTEST
        >>> import utool as ut
        >>> (out, err, ret) = ut.cmd('echo', 'hello world')
        >>> result = ut.repr4(list(zip(('out', 'err', 'ret'), (out, err, ret))), nobraces=True)
        >>> print(result)
        ('out', 'hello world\n'),
        ('err', None),
        ('ret', 0),

    Example1:
        >>> # ENABLE_DOCTEST
        >>> import utool as ut
        >>> target = ut.codeblock(
        ...      r'''
                 ('out', 'hello world\n'),
                 ('err', None),
                 ('ret', 0),
                 ''')
        >>> varydict = {
        ...    'shell': [True, False],
        ...    'detatch': [False],
        ...    'sudo': [True, False] if ut.get_argflag('--test-sudo') else [False],
        ...    'args': ['echo hello world', ('echo', 'hello world')],
        ... }
        >>> for count, kw in enumerate(ut.all_dict_combinations(varydict), start=1):
        >>>     print('+ --- TEST CMD %d ---' % (count,))
        >>>     print('testing cmd with params ' + ut.repr4(kw))
        >>>     args = kw.pop('args')
        >>>     restup = ut.cmd(args, pad_stdout=False, **kw)
        >>>     tupfields = ('out', 'err', 'ret')
        >>>     output = ut.repr4(list(zip(tupfields, restup)), nobraces=True)
        >>>     ut.assert_eq(output, target)
        >>>     print('L ___ TEST CMD %d ___\n' % (count,))

    Example2:
        >>> # DISABLE_DOCTEST
        >>> # UNSTABLE_DOCTEST
        >>> # ping is not as universal of a command as I thought
        >>> from utool.util_cplat import *  # NOQA
        >>> import utool as ut
        >>> varydict = {
        ...    'shell': [True, False],
        ...    'detatch': [True],
        ...    'args': ['ping localhost', ('ping', 'localhost')],
        ... }
        >>> proc_list = []
        >>> for count, kw in enumerate(ut.all_dict_combinations(varydict), start=1):
        >>>     print('+ --- TEST CMD %d ---' % (count,))
        >>>     print('testing cmd with params ' + ut.repr4(kw))
        >>>     args = kw.pop('args')
        >>>     restup = ut.cmd(args, pad_stdout=False, **kw)
        >>>     out, err, proc = restup
        >>>     proc_list.append(proc)
        >>>     print(proc)
        >>>     print(proc)
        >>>     print(proc.poll())
        >>>     print('L ___ TEST CMD %d ___\n' % (count,))
    """
    try:
        # Parse the keyword arguments
        verbose, detatch, shell, sudo, pad_stdout = __parse_cmd_kwargs(kwargs)
        quiet = kwargs.pop('quiet', False)
        silence = kwargs.pop('silence', False)
        if pad_stdout:
            sys.stdout.flush()
            print('\n+--------')
        args = __parse_cmd_args(args, sudo, shell)
        # Print what you are about to do
        if not quiet:
            print('[ut.cmd] RUNNING: %r' % (args,))
        # Open a subprocess with a pipe
        if kwargs.get('dryrun', False):
            print('[ut.cmd] Exiting because dryrun=True')
            return None, None, None
        proc = subprocess.Popen(args, stdout=subprocess.PIPE,
                                stderr=subprocess.STDOUT, shell=shell,
                                universal_newlines=True
                                # universal_newlines=False
                                )
        hack_use_stdout = True
        if detatch:
            if not quiet:
                print('[ut.cmd] PROCESS DETATCHING. No stdoutput can be reported...')
            # There is no immediate confirmation as to whether or not the script
            # finished. It might still be running for all you know
            return None, None, proc
        else:
            if verbose and not detatch:
                if not quiet:
                    print('[ut.cmd] RUNNING WITH VERBOSE OUTPUT')
                logged_out = []
                for line in _run_process(proc):
                    #line_ = line if six.PY2 else line.decode('utf-8')
                    line_ = line if six.PY2 else line
                    if len(line_) > 0:
                        if not silence:
                            if hack_use_stdout:
                                sys.stdout.write(line_)
                                sys.stdout.flush()
                            else:
                                # TODO make this play nicely with loggers
                                print_(line_)
                        logged_out.append(line)
                try:
                    from utool import util_str  # NOQA
                    # logged_out = util_str.ensure_unicode_strlist(logged_out)
                    out = '\n'.join(logged_out)
                except UnicodeDecodeError:
                    from utool import util_str  # NOQA
                    logged_out = util_str.ensure_unicode_strlist(logged_out)
                    out = '\n'.join(logged_out)
                    # print('logged_out = %r' % (logged_out,))
                    # raise
                (out_, err) = proc.communicate()
                #print('[ut.cmd] out: %s' % (out,))
                if not quiet:
                    try:
                        print('[ut.cmd] stdout: %s' % (out_,))
                        print('[ut.cmd] stderr: %s' % (err,))
                    except UnicodeDecodeError:
                        from utool import util_str  # NOQA
                        print('[ut.cmd] stdout: %s' % (util_str.ensure_unicode(out_),))
                        print('[ut.cmd] stderr: %s' % (util_str.ensure_unicode(err),))

            else:
                # Surpress output
                #print('[ut.cmd] RUNNING WITH SUPRESSED OUTPUT')
                (out, err) = proc.communicate()
            # Make sure process if finished
            ret = proc.wait()
            if not quiet:
                print('[ut.cmd] PROCESS FINISHED')
            if pad_stdout:
                print('L________\n')
            return out, err, ret
    except Exception as ex:
        import utool as ut
        #if isinstance(args, tuple):
        #    print(ut.truepath(args[0]))
        #elif isinstance(args, six.string_types):
        #    print(ut.unixpath(args))
        ut.printex(ex, 'Exception running ut.cmd',
                   keys=['verbose', 'detatch', 'shell', 'sudo', 'pad_stdout'],
                   tb=True)


def cmd2(command, shell=False, detatch=False, verbose=False, verbout=None):
    """
    Trying to clean up cmd

    Args:
        command (str): string command
        shell (bool): if True, process is run in shell
        detatch (bool): if True, process is run in background
        verbose (int): verbosity mode
        verbout (bool): if True, `command` writes to stdout in realtime.
            defaults to True iff verbose > 0

    Returns:
        dict: info - information about command status
    """
    import shlex
    if isinstance(command, (list, tuple)):
        raise ValueError('command tuple not supported yet')
    args = shlex.split(command, posix=not WIN32)
    if verbose is True:
        verbose = 2
    if verbout is None:
        verbout = verbose >= 1
    if verbose >= 2:
        print('+=== START CMD2 ===')
        print('Command:')
        print(command)
        if verbout:
            print('----')
            print('Stdout:')
    proc = subprocess.Popen(args, stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT, shell=shell,
                            universal_newlines=True)
    if detatch:
        info = {'proc': proc}
    else:
        write_fn = sys.stdout.write
        flush_fn = sys.stdout.flush
        logged_out = []
        for line in _run_process(proc):
            #line_ = line if six.PY2 else line.decode('utf-8')
            line_ = line if six.PY2 else line
            if len(line_) > 0:
                if verbout:
                    write_fn(line_)
                    flush_fn()
                logged_out.append(line)
        try:
            from utool import util_str  # NOQA
            # out = '\n'.join(logged_out)
            out = ''.join(logged_out)
        except UnicodeDecodeError:
            from utool import util_str  # NOQA
            logged_out = util_str.ensure_unicode_strlist(logged_out)
            # out = '\n'.join(logged_out)
            out = ''.join(logged_out)
            # print('logged_out = %r' % (logged_out,))
            # raise
        (out_, err) = proc.communicate()
        ret = proc.wait()
        info = {
            'out': out,
            'err': err,
            'ret': ret,
        }
    if verbose >= 2:
        print('L___ END CMD2 ___')
    return info


def get_flops():
    """ # DOESNT WORK """
    from sys import stdout
    from re import compile

    filename = "linpack.out"
    fpnum = r'\d+\.\d+E[+-]\d\d'
    fpnum_1 = fpnum + r' +'
    pattern = compile(r'^ *' + fpnum_1 + fpnum_1 + fpnum_1 + r'(' + fpnum + r') +' + fpnum_1 + fpnum + r' *\n$')
    speeds = [0.0, 1.0e75, 0.0]

    file = open(filename)
    count = 0
    while file :
        line = file.readline()
        if not line :
            break
        if pattern.match(line) :
            count = count + 1
            x = float(pattern.sub(r'\1', line))
            if x < 1.0 :
                print(count)
            speeds[0] = speeds[0] + x
            speeds[1] = min(speeds[1], x)
            speeds[2] = max(speeds[2], x)
    file.close()
    if count != 0 :
        speeds[0] = speeds[0] / count

    stdout.write("%6.1f MFlops (%d from %.1f to %.1f)\n" % (speeds[0], count, speeds[1], speeds[2]))


def set_process_title(title):
    try:
        import setproctitle
        setproctitle.setproctitle(title)
    except ImportError as ex:
        import utool
        utool.printex(ex, iswarning=True)


def is64bit_python():
    """
    Returns:
        True if running 64 bit python and False if running on 32 bit
    """
    #http://stackoverflow.com/questions/1405913/how-do-i-determine-if-my-python-shell-is-executing-in-32bit-or-64bit-mode-on-os
    is64bit = sys.maxsize > 2 ** 32
    #import platform
    #platform.architecture()
    #import ctypes
    #(ctypes.sizeof(ctypes.c_voidp))
    return is64bit


def get_python_dynlib():
    """
    python -c "import utool; print(utool.get_python_dynlib())"

    get_python_dynlib

    Returns:
        ?: dynlib

    Example:
        >>> # DISABLE_DOCTEST
        >>> # DOCTEST_DISABLE
        >>> from utool.util_cplat import *  # NOQA
        >>> dynlib = get_python_dynlib()
        >>> print(dynlib)
        /usr/lib/x86_64-linux-gnu/libpython2.7.so
    """
    import sysconfig
    cfgvars = sysconfig.get_config_vars()
    dynlib = os.path.join(cfgvars['LIBDIR'], cfgvars['MULTIARCH'], cfgvars['LDLIBRARY'])
    if not exists(dynlib):
        dynlib = os.path.join(cfgvars['LIBDIR'], cfgvars['LDLIBRARY'])
    assert exists(dynlib)
    return dynlib


def get_path_dirs():
    """
    returns a list of directories in the PATH system variable

    Returns:
        list: pathdirs

    CommandLine:
        python -m utool.util_cplat --exec-get_path_dirs

    Example:
        >>> # DISABLE_DOCTEST
        >>> # SCRIPT
        >>> from utool.util_cplat import *  # NOQA
        >>> import utool as ut
        >>> pathdirs = get_path_dirs()
        >>> result = ('pathdirs = %s' % (ut.repr4(pathdirs),))
        >>> print(result)

    """
    pathdirs = os.environ['PATH'].split(os.pathsep)
    return pathdirs


def print_path(sort=True):
    pathdirs = get_path_dirs()
    if sort:
        pathdirs = sorted(pathdirs)
    print('\n'.join(pathdirs))


def search_env_paths(fname, key_list=None, verbose=None):
    r"""
    Searches your PATH to see if fname exists

    Args:
        fname (str): file name to search for (can be glob pattern)

    CommandLine:
        python -m utool search_env_paths --fname msvcr*.dll
        python -m utool search_env_paths --fname '*flann*'

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_cplat import *  # NOQA
        >>> import utool as ut
        >>> fname = 'opencv2/highgui/libopencv_highgui.so'
        >>> fname = ut.get_argval('--fname', default='*')
        >>> print('fname = %r' % (fname,))
        >>> key_list = None # ['PATH']
        >>> found = search_env_paths(fname, key_list)
        >>> print(ut.repr4(found, nl=True, strvals=True))

    Ignore:
        OpenCV_DIR:PATH={share_opencv}
        OpenCV_CONFIG_PATH:FILEPATH={share_opencv}

    """
    import utool as ut
    # from os.path import join
    if key_list is None:
        key_list = [key for key in os.environ if key.find('PATH') > -1]
        print('key_list = %r' % (key_list,))

    found = ut.ddict(list)

    for key in key_list:
        dpath_list = os.environ[key].split(os.pathsep)
        for dpath in dpath_list:
            #if verbose:
            #    print('dpath = %r' % (dpath,))
            # testname = join(dpath, fname)
            matches = ut.glob(dpath, fname)
            found[key].extend(matches)
            #import fnmatch
            #import utool
            #utool.embed()
            #if ut.checkpath(testname, verbose=False):
            #    if verbose:
            #        print('Found in key=%r' % (key,))
            #        ut.checkpath(testname, verbose=True, info=True)
            #    found += [testname]
    return dict(found)


def __debug_win_msvcr():
    import utool as ut
    fname = 'msvcr*.dll'
    key_list = ['PATH']
    found = ut.search_env_paths(fname, key_list)
    fpaths = ut.unique(ut.flatten(found.values()))
    fpaths = ut.lmap(ut.ensure_unixslash, fpaths)
    from os.path import basename
    dllnames = [basename(x) for x in fpaths]
    grouped = dict(ut.group_items(fpaths, dllnames))
    print(ut.repr4(grouped, nl=4))

    keytoid = {
    }

    for key, vals in grouped.items():
        infos = ut.lmap(ut.get_file_nBytes, vals)
        #infos = ut.lmap(ut.get_file_uuid, vals)
        #uuids = [ut.get_file_uuid(val) for val in vals]
        keytoid[key] = list(zip(infos, vals))
    ut.print_dict(keytoid, nl=2)


def change_term_title(title):
    r"""
    only works on unix systems only tested on Ubuntu GNOME changes text on
    terminal title for identifying debugging tasks.

    The title will remain until python exists

    Args:
        title (str):

    References:
        http://stackoverflow.com/questions/5343265/setting-title-for-tabs-in-terminator-console-application-in-ubuntu/8850484#8850484

    CommandLine:
        python -m utool change_term_title
        echo -en "\033]0;newtitle\a"

         printf "\e]2;newtitle\a";

        echo -en "\033]0;DocTest /home/joncrall/code/ibeis/ibeis.algo.graph.core.py --test-AnnotInference._make_state_delta\a"

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_cplat import *  # NOQA
        >>> title = 'change title test'
        >>> result = change_term_title(title)
        >>> print(result)
    """
    if True:
        # Disabled
        return
    if not WIN32:
        #print("CHANGE TERM TITLE to %r" % (title,))
        if title:
            #os.environ['PS1'] = os.environ['PS1'] + '''"\e]2;\"''' + title + '''\"\a"'''
            cmd_str = r'''echo -en "\033]0;''' + title + '''\a"'''
            os.system(cmd_str)


def send_keyboard_input(text=None, key_list=None):
    """
    Args:
        text (None):
        key_list (list):

    References:
        http://stackoverflow.com/questions/14788036/python-win32api-sendmesage
        http://www.pinvoke.net/default.aspx/user32.sendinput

    CommandLine:
        python -m utool.util_cplat --test-send_keyboard_input

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_cplat import *  # NOQA
        >>> text = '%paste'
        >>> result = send_keyboard_input('%paste')
        >>> print(result)
    """
    #key_mapping = {
    #    'enter':
    #}
    if WIN32:
        #raise NotImplementedError()
        #import win32api
        #import win32gui
        #import win32con
        #hwnd = win32gui.GetForegroundWindow()
        #print('entering text into %r' % (win32gui.GetWindowText(hwnd ),))
        #win32con.VK_RETURN

        #def callback(hwnd, hwnds):
        #   #if win32gui.IsWindowVisible(hwnd) and win32gui.IsWindowEnabled(hwnd):
        #       #hwnds[win32gui.GetClassName(hwnd)] = hwnd
        #   #return True
        #hwnds = {}
        #win32gui.EnumChildWindows(hwnd, callback, hwnds)

        #for ord_char in map(ord, text):
        #   #win32api.SendMessage(hwnd, win32con.WM_CHAR, ord_char, 0)
        from utool._internal import win32_send_keys
        pause = float(.05)
        text = 'paste'
        keys = text
        kw = dict(with_spaces=False, with_tabs=True, with_newlines=False)
        win32_send_keys.SendKeys(keys, pause=pause, turn_off_numlock=True, **kw)
        #win32_send_keys
        #import time
        #keys_ = win32_send_keys.parse_keys(keys, **kw)
        #for k in keys_:
        #    k.Run()
        #    time.sleep(pause)

    else:
        if key_list is None:
            char_map = {
                '%': 'shift+5'
            }
            key_list = [char_map.get(char, char) for char in text]
        xdotool_args = ['xdotool', 'key'] + key_list
        #, 'shift+5', 'p', 'a', 's', 't', 'e', 'enter']
        cmd = ' '.join(xdotool_args)
        print('Running: cmd=%r' % (cmd,))
        print('+---')
        print(cmd)
        print('L___')
        os.system(cmd)


def ipython_paste(*args, **kwargs):
    """ pastes for me FIXME: make something like this work on unix and windows"""
    if WIN32:
        pass
    else:
        winhandle = 'joncrall@Hyrule'
        args1 = ['wmctrl', '-a', winhandle]
        args2 = ['xdotool', 'key', 'shift+5', 'p', 'a', 's', 't', 'e', 'enter']
        os.system(' '.join(args1))
        os.system(' '.join(args2))

    #fallback_execute(args1)
    #fallback_execute(args2)


def spawn_delayed_ipython_paste():
    import utool as ut
    # Gonna be pasting
    def delayed_ipython_paste(delay):
        import time
        import utool as ut
        #import os
        print('waiting')
        time.sleep(delay)
        ut.send_keyboard_input(text='%paste')
        ut.send_keyboard_input(key_list=['KP_Enter'])
        #os.system(' '.join(['xdotool', 'key', 'shift+5', 'p', 'a', 's', 't', 'e', 'KP_Enter']))
    ut.spawn_background_thread(delayed_ipython_paste, .1)


def print_system_users():
    r"""

    prints users on the system

    On unix looks for /bin/bash users in /etc/passwd

    CommandLine:
        python -m utool.util_cplat --test-print_system_users

    Example:
        >>> # DISABLE_DOCTEST
        >>> # SCRIPT
        >>> from utool.util_cplat import *  # NOQA
        >>> result = print_system_users()
        >>> print(result)
    """
    import utool as ut
    text = ut.read_from('/etc/passwd')
    userinfo_text_list = text.splitlines()
    userinfo_list = [uitext.split(':') for uitext in userinfo_text_list]
    #print(ut.repr4(sorted(userinfo_list)))
    bash_users = [tup for tup in userinfo_list if tup[-1] == '/bin/bash']
    print(ut.repr4(sorted(bash_users)))


def check_installed_debian(pkgname):
    """
    References:
        http://www.cyberciti.biz/faq/find-out-if-package-is-installed-in-linux/
    """
    import utool as ut
    #pkgname = 'espeak'
    #pkgname = 'sudo'
    #ut.cmd('hash ' + pkgname + ' 2>/dev/null')
    tup = ut.cmd('hash ' + pkgname + ' 2>/dev/null', quiet=True, pad_stdout=False)
    out, err, ret = tup
    is_installed = (ret == 0)
    return is_installed


def assert_installed_debian(pkgname):
    import utool as ut
    if not ut.check_installed_debian(pkgname):
        raise AssertionError('espeak must be installed. run sudo apt-get install -y ' + pkgname)


def unload_module(modname):
    """
    WARNING POTENTIALLY DANGEROUS AND MAY NOT WORK

    References:
        http://stackoverflow.com/questions/437589/how-do-i-unload-reload-a-python-module

    CommandLine:
        python -m utool.util_cplat --test-unload_module

    Example:
        >>> # DISABLE_DOCTEST
        >>> import sys, gc  # NOQA
        >>> import pyhesaff
        >>> import utool as ut
        >>> modname = 'pyhesaff'
        >>> print('%s refcount=%r' % (modname, sys.getrefcount(pyhesaff),))
        >>> #referrer_list = gc.get_referrers(sys.modules[modname])
        >>> #print('referrer_list = %s' % (ut.repr4(referrer_list),))
        >>> ut.unload_module(modname)
        >>> assert pyhesaff is None

    """
    import sys
    import gc
    if modname in sys.modules:
        referrer_list = gc.get_referrers(sys.modules[modname])
        #module = sys.modules[modname]
        for referer in referrer_list:
            if referer is not sys.modules:
                referer[modname] = None
            #del referer[modname]
        #sys.modules[modname] = module
        #del module
        refcount = sys.getrefcount(sys.modules[modname])
        print('%s refcount=%r' % (modname, refcount))
        del sys.modules[modname]


#def get_ipython_config_file():
#    """
#    or to create an empty default profile, populated with default config files:
#    ipython profile create
#    """

#from subprocess import check_output
#http://stackoverflow.com/questions/8015163/how-to-check-screen-is-running
#def screen_present(name):
#        var = check_output(["screen -ls; true"],shell=True)
#        if "."+name+"\t(" in var:
#                print name+" is running"
#        else:
#                print name+" is not running"


def pip_install(package):
    """
    References:
        http://stackoverflow.com/questions/15974100/ipython-install-new-modules
    """
    import pip
    pip.main(['install', package])


if __name__ == '__main__':
    """
    CommandLine:
        python -c "import utool, utool.util_cplat; utool.doctest_funcs(utool.util_cplat, allexamples=True)"
        python -c "import utool, utool.util_cplat; utool.doctest_funcs(utool.util_cplat)"
        python -m utool.util_cplat
        python -m utool.util_cplat --allexamples
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
"""
util_config

TODO: FINISH ME AND ALLOW FOR CUSTOM SETTINGS BASED OFF OF A USER PROFILE
"""
from __future__ import absolute_import, division, print_function
from utool import util_inject
(print, print_, printDBG, rrr, profile) = util_inject.inject(__name__, '[config]')


def read_repo_config():
    pass


def get_default_repo_config():
    """
    import utool
    """
    REPO_CONFIG = {
        'project_name': None,
        'authors': [],
        'licence': None,
        'enable_cyth': None,
        'docstr_style': None,
    }

    return REPO_CONFIG


def get_default_global_config():
    GLOBAL_CONFIG = {
        # Supplementary Info
        'AUTHORS': {
            'username': {'name': None,
                         'email': None, },

            'kingarthur': {'name': 'Arthur, King of the Britans',
                           'email': 'kingarthur@camelot.com',
                           'permited repos': ['kotr'],
                           'aliases': [],
                           'machines': [],
                           'quest': 'To seek the holy grail.',
                           'sa5ws5%lp&30sjnk': 'What to you mean? An African or European swallow?', },

            'joncrall': {'name': 'Jon Crall',
                         'email': 'erotemic@gmail.com',
                         'permited repos': ['ibeis', 'utool', 'hesaff',
                                            'guitool', 'plottool', 'vtool'],
                         'aliases': [],
                         'machines': [], },

        },
    }
    return GLOBAL_CONFIG


def write_default_repo_config():
    import utool
    CONFIG_DICT = utool.get_default_repo_config()
    config_str = utool.repr2(CONFIG_DICT, strvals=True, newlines=True,
                             recursive=True)
    print(config_str)


if __name__ == '__main__':
    """
    CommandLine:
        python -m utool.util_config
        python -m utool.util_config --allexamples
        python -m utool.util_config --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function, unicode_literals
import os
from utool import util_inject
from utool import util_str
print, rrr, profile = util_inject.inject2(__name__)

try:
    # Resource does not exist in win32
    import resource

    def time_in_usermode():
        stime = resource.getrusage(resource.RUSAGE_SELF).ru_stime
        return stime

    def time_in_systemmode():
        utime = resource.getrusage(resource.RUSAGE_SELF).ru_utime
        return utime

    def peak_memory():
        """Returns the resident set size (the portion of
        a process's memory that is held in RAM.)
        """
        # MAXRSS is expressed in kilobytes. Convert to bytes
        # FIXME: MAXRSS is NOT expressed in kilobytes. use resource.getpagesize to Convert to bytes
        # References: http://stackoverflow.com/questions/938733/total-memory-used-by-python-process
        #resource.getpagesize
        maxrss = resource.getrusage(resource.RUSAGE_SELF).ru_maxrss * 1024
        return maxrss

    def get_resource_limits():
        #rlimit_keys = [key for key in six.iterkeys(resource.__dict__) if key.startswith('RLIMIT_')]
        #print('\n'.join(['(\'%s\', resource.%s),' % (key.replace('RLIMIT_', ''), key) for key in rlimit_keys]))
        rlim_keytups = [
            ('MEMLOCK', resource.RLIMIT_MEMLOCK),
            ('NOFILE', resource.RLIMIT_NOFILE),
            ('CPU', resource.RLIMIT_CPU),
            ('DATA', resource.RLIMIT_DATA),
            ('OFILE', resource.RLIMIT_OFILE),
            ('STACK', resource.RLIMIT_STACK),
            ('FSIZE', resource.RLIMIT_FSIZE),
            ('CORE', resource.RLIMIT_CORE),
            ('NPROC', resource.RLIMIT_NPROC),
            ('AS', resource.RLIMIT_AS),
            ('RSS', resource.RLIMIT_RSS),
        ]
        rlim_valtups = [(lbl, resource.getrlimit(rlim_key)) for (lbl, rlim_key) in rlim_keytups]
        def rlimval_str(rlim_val):
            soft, hard = rlim_val
            softstr = util_str.byte_str2(soft) if soft != -1 else 'None'
            hardstr = util_str.byte_str2(hard) if hard != -1 else 'None'
            return '%12s, %12s' % (softstr, hardstr)
        rlim_strs = ['%8s: %s' % (lbl, rlimval_str(rlim_val)) for (lbl, rlim_val) in rlim_valtups]
        print('Resource Limits: ')
        print('%8s  %12s  %12s' % ('id', 'soft', 'hard'))
        print('\n'.join(rlim_strs))
        return rlim_strs

    #def rusage_flags():
        #0	ru_utime	time in user mode (float)
        #1	ru_stime	time in system mode (float)
        #2	ru_maxrss	maximum resident set size
        #3	ru_ixrss	shared memory size
        #4	ru_idrss	unshared memory size
        #5	ru_isrss	unshared stack size
        #6	ru_minflt	page faults not requiring I/O
        #7	ru_majflt	page faults requiring I/O
        #8	ru_nswap	number of swap outs
        #9	ru_inblock	block input operations
        #10	ru_oublock	block output operations
        #11	ru_msgsnd	messages sent
        #12	ru_msgrcv	messages received
        #13	ru_nsignals	signals received
        #14	ru_nvcsw	voluntary context switches
        #15	ru_nivcsw	involuntary context switches
except ImportError:
    def time_in_usermode():
        raise NotImplementedError('unavailable in win32')

    def time_in_systemmode():
        raise NotImplementedError('unavailable in win32')

    def peak_memory():
        """Returns the resident set size (the portion of
        a process's memory that is held in RAM.)
        """
        raise NotImplementedError('unavailable in win32')

    def get_resource_limits():
        raise NotImplementedError('unavailable in win32')


def time_str2(seconds):
    return '%.2f sec' % (seconds,)


def print_resource_usage():
    print(get_resource_usage_str())


def get_resource_usage_str():
    usage_str_list = [
        ('+______________________'),
        ('|    RESOURCE_USAGE    process.get_memory_info()[0] / float(2 ** 20) '),
        ('|  * current_memory = %s' % util_str.byte_str2(current_memory_usage())),
    ]
    try:
        usage_str_list.extend([
            ('|  * peak_memory    = %s' % util_str.byte_str2(peak_memory())),
            ('|  * user_time      = %s' % time_str2(time_in_usermode())),
            ('|  * system_time    = %s' % time_str2(time_in_systemmode())),
        ])
    except Exception:
        pass
    usage_str_list.append('L______________________')
    usage_str = '\n'.join(usage_str_list)
    return usage_str


def current_memory_usage():
    """
    Returns this programs current memory usage in bytes
    """
    import psutil
    proc = psutil.Process(os.getpid())
    #meminfo = proc.get_memory_info()
    meminfo = proc.memory_info()
    rss = meminfo[0]  # Resident Set Size / Mem Usage
    vms = meminfo[1]  # Virtual Memory Size / VM Size  # NOQA
    return rss


def get_matching_process_ids(cmd_pattern, user_pattern):
    """
    CommandLine:
        export PID=30196
        export PID=$(python -c "import utool as ut; print(ut.get_matching_process_ids('jonc', 'python2.7'))")
        export PID=$(python -c "import utool as ut; print(ut.get_matching_process_ids('jonc', 'matlab'))")
        sudo -H echo $PID
        ps -o pid,comm,nice -p $PID
        renice 10 -p $PID
        sudo renice -4 -p $PID

    user_pattern = 'jonc'
    cmd_pattern = 'main.py'
    user_pattern = None
    cmd_pattern = 'matlab'
    get_matching_process_ids(cmd_pattern, user_pattern)
    """
    import psutil
    import re
    process_list = list(psutil.process_iter())
    def matches_pattern(proc, user_pattern, cmd_pattern):
        matches_user = True if user_pattern is None else re.match(user_pattern, proc.username())
        cmdline_str = ' '.join(proc.cmdline())
        matches_name = True if cmd_pattern is None else re.search(cmd_pattern, cmdline_str)
        return matches_user and matches_name
    filtered_proc_list = [proc for proc in process_list if matches_pattern(proc, user_pattern, cmd_pattern)]

    for proc in filtered_proc_list:
        print(' | '.join([str(proc.username()), str(proc.nice()), str(proc), ' '.join(proc.cmdline())]))
        #print(proc.cmdline())
        #print(proc.pid)
        #print('---')

    important_process_list = [proc for proc in process_list if proc.nice() < -4]
    for proc in important_process_list:
        print(' -- '.join([str(proc.username()), str(proc.nice()), str(proc), ' '.join(proc.cmdline())]))

    #for proc in filtered_proc_list:
    #    print('---')
    #    print(proc)
    #    print(proc.cmdline())
    #    print(proc.nice())
    #    print(proc.pid)
    filtered_pid_list = [proc.pid for proc in filtered_proc_list]
    return filtered_pid_list


def num_cpus():
    import psutil
    return psutil.cpu_count(logical=True)


def num_unused_cpus(thresh=10):
    """
    Returns the number of cpus with utilization less than `thresh` percent
    """
    import psutil
    cpu_usage = psutil.cpu_percent(percpu=True)
    return sum([p < thresh for p in cpu_usage])


def available_memory():
    """
    Returns total system wide available memory in bytes
    """
    import psutil
    return psutil.virtual_memory().available


def total_memory():
    """
    Returns total system wide memory in bytes
    """
    import psutil
    return psutil.virtual_memory().total


def used_memory():
    """
    Returns total system wide used memory in bytes
    """
    return total_memory() - available_memory()


def memstats():
    print(get_memstats_str())


def get_memstats_str():
    return '\n'.join([
        ('[psutil] total     = %s' % util_str.byte_str2(total_memory())),
        ('[psutil] available = %s' % util_str.byte_str2(available_memory())),
        ('[psutil] used      = %s' % util_str.byte_str2(used_memory())),
        ('[psutil] current   = %s' % util_str.byte_str2(current_memory_usage())),
    ])


def get_python_datastructure_sizes():
    """
    References:
        http://stackoverflow.com/questions/1331471/in-memory-size-of-python-stucture

    CommandLine:
        python -m utool.util_resources --test-get_python_datastructure_sizes

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_resources import *  # NOQA
        >>> import utool as ut  # NOQA
        >>> type_sizes = get_python_datastructure_sizes()
        >>> result = ut.repr4(type_sizes, sorted_=True)
        >>> print(result)
    """
    import sys
    import decimal
    import six

    empty_types = {
        'int'     : 0,
        'float'   : 0.0,
        'dict'    : dict(),
        'set'     : set(),
        'tuple'   : tuple(),
        'list'    : list(),
        'str'     : '',
        'unicode' : u'',
        'decimal' : decimal.Decimal(0),
        'object'  : object(),
    }
    type_sizes = {key: sys.getsizeof(val)
                  for key, val in six.iteritems(empty_types)}
    return type_sizes

#psutil.virtual_memory()
#psutil.swap_memory()
#psutil.disk_partitions()
#psutil.disk_usage("/")
#psutil.disk_io_counters()
#psutil.net_io_counters(pernic=True)
#psutil.get_users()
#psutil.get_boot_time()
#psutil.get_pid_list()

if __name__ == '__main__':
    """
    CommandLine:
        python -m utool.util_resources
        python -m utool.util_resources --allexamples
        python -m utool.util_resources --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
"""
Injects code into live modules or into text source files.

Basic use case is to extend the print function into a logging function
"""
from __future__ import absolute_import, division, print_function, unicode_literals
from six.moves import builtins, range, zip, map  # NOQA
import six  # NOQA
import sys
import functools
from utool._internal import meta_util_six
from utool._internal import meta_util_arg
from utool import util_logging
import traceback


__AGGROFLUSH__ = '--aggroflush' in sys.argv
__LOGGING__    = '--logging'    in sys.argv
__DEBUG_ALL__  = '--debug-all'  in sys.argv
__DEBUG_PROF__ = '--debug-prof' in sys.argv or '--debug-profile' in sys.argv
DEBUG_PRINT = '--debug-print' in sys.argv
DEBUG_PRINT_N = meta_util_arg.get_argval('--debug-print-N', type_=str, default=None)
if DEBUG_PRINT_N is not None:
    DEBUG_PRINT_N = list(map(int, DEBUG_PRINT_N.split(',')))
    DEBUG_PRINT = True
elif DEBUG_PRINT:
    DEBUG_PRINT_N = 0

QUIET = '--quiet' in sys.argv
SILENT = '--silent' in sys.argv
VERYVERBOSE = meta_util_arg.VERYVERBOSE
VERBOSE = meta_util_arg.VERBOSE
PRINT_INJECT_ORDER = meta_util_arg.PRINT_INJECT_ORDER  # --verbinject
# only word
EXIT_ON_INJECT_MODNAME = meta_util_arg.get_argval('--exit-on-inject', type_=str, default=None)


if __LOGGING__:
    util_logging.start_logging()

# Read all flags with --debug in them
ARGV_DEBUG_FLAGS = []
for argv in sys.argv:
    if argv.startswith('--debug-'):
        ARGV_DEBUG_FLAGS.append(argv.replace('--debug-', '').replace('-', '_'))


#print('ARGV_DEBUG_FLAGS: %r' % (ARGV_DEBUG_FLAGS,))

#__STDOUT__ = sys.stdout
#__PRINT_FUNC__     = builtins.print
#__PRINT_DBG_FUNC__ = builtins.print
#__WRITE_FUNC__ = __STDOUT__.write
#__FLUSH_FUNC__ = __STDOUT__.flush
__RELOAD_OK__  = '--noreloadable' not in sys.argv


__INJECTED_MODULES__ = set([])

# Do not inject into these modules
__INJECT_BLACKLIST__ = frozenset(['tri', 'gc', 'sys', 'string', 'types',
                                  '_dia', 'responce', 'six', __name__])


def _inject_funcs(module, *func_list):
    for func in func_list:
        if (module is not None and
                hasattr(module, '__name__') and
                module.__name__ not in __INJECT_BLACKLIST__ and
                not module.__name__.startswith('six') and
                not module.__name__.startswith('sys')):
            #print('setting: %s.%s = %r' % (module.__name__, meta_util_six.get_funcname(func), func))
            setattr(module, meta_util_six.get_funcname(func), func)


def _add_injected_module(module):
    global __INJECTED_MODULES__
    __INJECTED_MODULES__.add(module)


def get_injected_modules():
    return list(__INJECTED_MODULES__)


def _get_module(module_name=None, module=None, register=True):
    """ finds module in sys.modules based on module name unless the module has
    already been found and is passed in """
    if module is None and module_name is not None:
        try:
            module = sys.modules[module_name]
        except KeyError as ex:
            print(ex)
            raise KeyError(('module_name=%r must be loaded before ' +
                            'receiving injections') % module_name)
    elif module is not None and module_name is None:
        pass
    else:
        raise ValueError('module_name or module must be exclusively specified')
    if register is True:
        _add_injected_module(module)
    return module


def colored_pygments_excepthook(type_, value, tb):
    """
    References:
        https://stackoverflow.com/questions/14775916/color-exceptions-python

    CommandLine:
        python -m utool.util_inject --test-colored_pygments_excepthook

    """
    tbtext = ''.join(traceback.format_exception(type_, value, tb))
    try:
        from utool import util_str
        formatted_text = util_str.highlight_text(tbtext, lexer_name='pytb',
                                                 stripall=True)
    except Exception:
        # FIXME silent errro
        formatted_text = tbtext
        return sys.__excepthook__(type_, value, tb)
        #import utool as ut
        #if ut.SUPER_STRICT:
        #    raise
    sys.stderr.write(formatted_text)

    #EMBED_ON_ERROR = True
    # Doesn't work
    #if EMBED_ON_ERROR:
    #    import utool as ut
    #    ut.embed(N=1)


def inject_colored_exceptions():
    """
    Causes exceptions to be colored if not already

    Hooks into sys.excepthook

    CommandLine:
        python -m utool.util_inject --test-inject_colored_exceptions

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_inject import *  # NOQA
        >>> print('sys.excepthook = %r ' % (sys.excepthook,))
        >>> #assert sys.excepthook is colored_pygments_excepthook, 'bad excepthook'
        >>> raise Exception('should be in color')

    """
    #COLORED_INJECTS = '--nocolorex' not in sys.argv
    #COLORED_INJECTS = '--colorex' in sys.argv
    # Ignore colored exceptions on win32
    if VERBOSE:
        print('[inject] injecting colored exceptions')
    if not sys.platform.startswith('win32'):
        if VERYVERBOSE:
            print('[inject] injecting colored exceptions')
        if '--noinject-color' in sys.argv:
            print('Not injecting color')
        else:
            sys.excepthook = colored_pygments_excepthook
    else:
        if VERYVERBOSE:
            print('[inject] cannot inject colored exceptions')


def make_module_print_func(module):
    if SILENT:
        def print(*args, **kwargs):
            """ silent builtins.print """
            pass
    else:
        if DEBUG_PRINT:
            # Turns on printing where a message came from
            def print(*args, **kwargs):
                """ debugging logging builtins.print """
                from utool._internal.meta_util_dbg import get_caller_name
                calltag = ''.join(('[caller:', get_caller_name(N=DEBUG_PRINT_N), ']' ))
                util_logging._utool_print()(calltag, *args, **kwargs)
        else:
            def print(*args, **kwargs):
                """ logging builtins.print """
                util_logging._utool_print()(*args, **kwargs)
    return print


def make_module_write_func(module):
    if SILENT:
        def print_(*args, **kwargs):
            """ silent stdout.write """
            pass
    else:
        if __AGGROFLUSH__:
            def print_(*args, **kwargs):
                """ aggressive logging stdout.write """
                util_logging._utool_write()(*args)
                util_logging._utool_flush()()
        else:
            def print_(*args, **kwargs):
                """ logging stdout.write """
                util_logging._utool_write()(*args)
    return print_


def inject_print_functions(module_name=None, module_prefix='[???]',
                           DEBUG=False, module=None):
    """
    makes print functions to be injected into the module
    """
    module = _get_module(module_name, module)
    if SILENT:
        def print(*args):
            """ silent builtins.print """
            pass
        def printDBG(*args):
            """ silent debug print """
            pass
        def print_(*args):
            """ silent stdout.write """
            pass
    else:
        if DEBUG_PRINT:
            # Turns on printing where a message came from
            def print(*args):
                """ debugging logging builtins.print """
                from utool._internal.meta_util_dbg import get_caller_name
                calltag = ''.join(('[caller:', get_caller_name(N=DEBUG_PRINT_N), ']' ))
                util_logging._utool_print()(calltag, *args)
        else:
            def print(*args):
                """ logging builtins.print """
                util_logging._utool_print()(*args)

        if __AGGROFLUSH__:
            def print_(*args):
                """ aggressive logging stdout.write """
                util_logging._utool_write()(*args)
                util_logging._utool_flush()()
        else:
            def print_(*args):
                """ logging stdout.write """
                util_logging._utool_write()(*args)

        # turn on module debugging with command line flags
        dotpos = module.__name__.rfind('.')
        if dotpos == -1:
            module_name = module.__name__
        else:
            module_name = module.__name__[dotpos + 1:]
        def _replchars(str_):
            return str_.replace('_', '-').replace(']', '').replace('[', '')
        flag1 = '--debug-%s' % _replchars(module_name)
        flag2 = '--debug-%s' % _replchars(module_prefix)
        DEBUG_FLAG = any([flag in sys.argv for flag in [flag1, flag2]])
        for curflag in ARGV_DEBUG_FLAGS:
            if curflag in module_prefix:
                DEBUG_FLAG = True
        if __DEBUG_ALL__ or DEBUG or DEBUG_FLAG:
            print('INJECT_PRINT: %r == %r' % (module_name, module_prefix))
            def printDBG(*args):
                """ debug logging print """
                msg = ', '.join(map(str, args))
                util_logging.__UTOOL_PRINTDBG__(module_prefix + ' DEBUG ' + msg)
        else:
            def printDBG(*args):
                """ silent debug logging print """
                pass
    #_inject_funcs(module, print, print_, printDBG)
    print_funcs = (print, print_, printDBG)
    return print_funcs


def reload_module(module, verbose=None):
    if not __RELOAD_OK__:
        raise Exception('Reloading has been forced off')
    if verbose is None:
        verbose = 0 if QUIET else 1
    try:
        v = sys.version_info
        if v.major >= 3 and v.minor >= 4:
            import importlib
            reload = importlib.reload
        else:
            import imp
            reload = imp.reload
        if verbose:
            module_name = getattr(module, '__name__', '???')
            builtins.print('RELOAD: module __name__=' + module_name)
        reload(module)
    except Exception as ex:
        print(ex)
        print('[util_inject] Failed to reload %r' % (module,))
        raise


def make_module_reload_func(module_name=None, module_prefix='[???]', module=None):
    """ Injects dynamic module reloading """
    module = _get_module(module_name, module, register=False)
    if module_name is None:
        module_name = str(module.__name__)
    def rrr(verbose=True):
        """ Dynamic module reloading """
        if not __RELOAD_OK__:
            raise Exception('Reloading has been forced off')
        try:
            import imp
            if verbose and not QUIET:
                builtins.print('RELOAD: ' + str(module_prefix) + ' __name__=' + module_name)
            imp.reload(module)
        except Exception as ex:
            print(ex)
            print('%s Failed to reload' % module_prefix)
            raise
    # this doesn't seem to set anything on import *
    #_inject_funcs(module, rrr)
    return rrr


def DUMMYPROF_FUNC(func):
    """ dummy profiling func. does nothing """
    return func


def TIMERPROF_FUNC(func):
    @functools.wraps(func)
    def prof_wrapper(*args, **kwargs):
        import utool as ut
        with ut.Timer(meta_util_six.get_funcname(func)):
            return func(*args, **kwargs)
        #return ret
    return prof_wrapper


if '--profile' in sys.argv:
    #util_profile.make_profiler()
    import line_profiler
    PROFILE_FUNC = line_profiler.LineProfiler()
    PROFILING = True
    if __DEBUG_PROF__:
        print('[util_inject] PROFILE ON')
else:
    PROFILING = False
    PROFILE_FUNC = DUMMYPROF_FUNC
    #PROFILE_FUNC = TIMERPROF_FUNC
    if __DEBUG_PROF__:
        print('[util_inject] PROFILE OFF')


# Look in command line for functions to profile
PROF_FUNC_PAT_LIST = meta_util_arg.get_argval('--prof-func', type_=str, default=None)
if PROF_FUNC_PAT_LIST is not None:
    PROF_FUNC_PAT_LIST = PROF_FUNC_PAT_LIST.split(',')
    print('[util_inject] PROF_FUNC_PAT_LIST: %r' % (PROF_FUNC_PAT_LIST,))

# Look in command line for modules to profile
PROF_MOD_PAT_LIST = meta_util_arg.get_argval('--prof-mod', type_=str, default=None)
if PROF_MOD_PAT_LIST is not None:
    PROF_MOD_PAT_LIST = PROF_MOD_PAT_LIST.split(',')
    print('[util_inject] PROF_MOD_PAT_LIST: %r' % (PROF_MOD_PAT_LIST,))


def memprof(func):
    """ requires memory_profiler
    pip install memory_profiler

    References:
        https://pypi.python.org/pypi/memory_profiler

    """
    import memory_profiler
    return memory_profiler.profile(func)


def _matches_list(name, pat_list):
    return any([name.find(pat) != -1 for pat in pat_list])


def _profile_func_flag(funcname):
    """ checks if func has been requested to be profiled """
    if PROF_FUNC_PAT_LIST is None:
        return True
    return _matches_list(funcname, PROF_FUNC_PAT_LIST)


def _profile_module_flag(module_name):
    """ checks if module has been requested to be profiled """
    if PROF_MOD_PAT_LIST is None:
        return True
    return _matches_list(module_name, PROF_MOD_PAT_LIST)


def make_module_profile_func(module_name=None, module_prefix='[???]', module=None):
    # FIXME: not injecting right
    module = _get_module(module_name, module)
    if not _profile_module_flag(str(module)):
        return DUMMYPROF_FUNC
    #if module_name is None:
    #    return DUMMYPROF_FUNC
    #profile_module_flag = PROF_MODULE_PAT is None or module_name.startswith(PROF_MODULE_PAT)
    #if not profile_module_flag:
    #    return DUMMYPROF_FUNC

    def profile_withfuncname_filter(func):
        # Test to see if this function is specified
        funcname = meta_util_six.get_funcname(func)
        if _profile_func_flag(funcname):
            if __DEBUG_PROF__:
                print('profile func %r' % (func,))
            # if isinstance(func, six.class_types):
            #     for k in func.__dict__.keys():
            #         if k.startswith('_'):
            #             continue
            #         v = getattr(func, k)
            #         if str(type(v)) == 'function':
            #             setattr(func, k, PROFILE_FUNC(v))
            # else:
            return PROFILE_FUNC(func)
        return func
    return profile_withfuncname_filter


DEBUG_SLOW_IMPORT = False
if DEBUG_SLOW_IMPORT:
    # Find which modules take the longest to import
    import ubelt as ub
    tt = ub.Timer(verbose=False)
    tt.tic()
    import_times = {}
    PREV_MODNAME = None

    def check_debug_import_times():
        import utool as ut
        from utool import util_inject
        print(ut.align(ut.repr4(ut.sort_dict(util_inject.import_times, 'vals'), precision=4), ':'))
        # ututil_inject.import_times
        # pass


def noinject(module_name=None, module_prefix='[???]', DEBUG=False, module=None, N=0, via=None):
    """
    Use in modules that do not have inject in them

    Does not inject anything into the module. Just lets utool know that a module
    is being imported so the import order can be debuged
    """
    if PRINT_INJECT_ORDER:
        from utool._internal import meta_util_dbg
        callername = meta_util_dbg.get_caller_name(N=N + 1, strict=False)
        lineno = meta_util_dbg.get_caller_lineno(N=N + 1, strict=False)
        suff = ' via %s' % (via,) if via else ''
        fmtdict = dict(N=N, lineno=lineno, callername=callername,
                       modname=module_name, suff=suff)
        msg = '[util_inject] N={N} {modname} is imported by {callername} at lineno={lineno}{suff}'.format(**fmtdict)

        if DEBUG_SLOW_IMPORT:
            global PREV_MODNAME
            seconds = tt.toc()
            import_times[(PREV_MODNAME, module_name)] = seconds
            PREV_MODNAME = module_name

        builtins.print(msg)

        if DEBUG_SLOW_IMPORT:
            tt.tic()
        # builtins.print(elapsed)
        if EXIT_ON_INJECT_MODNAME == module_name:
            builtins.print('...exiting')
            assert False, 'exit in inject requested'


def inject(module_name=None, module_prefix='[???]', DEBUG=False, module=None, N=1):
    """
    Injects your module with utool magic

    Utool magic is not actually magic. It just turns your ``print`` statments into
    logging statments, allows for your module to be used with the utool.Indent
    context manager and the and utool.indent_func decorator. ``printDBG`` will soon
    be deprecated as will ``print_``. The function rrr is a developer convinience for
    reloading the module dynamically durring runtime. The profile decorator is
    a no-op if not using kernprof.py, otherwise it is kernprof.py's profile
    decorator.

    Args:
        module_name (str): the __name__ varaible in your module
        module_prefix (str): a user defined module prefix
        DEBUG (bool):
        module (None): the actual module (optional)

    Returns:
        tuple : (print, print_, printDBG, rrr, profile_)

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_inject import *  # NOQA
        >>> from __future__ import absolute_import, division, print_function, unicode_literals
        >>> from util.util_inject import inject
        >>> print, rrr, profile = inject2(__name__, '[mod]')
    """
    #noinject(module_name, module_prefix, DEBUG, module, N=1)
    noinject(module_name, module_prefix, DEBUG, module, N=N)
    module = _get_module(module_name, module)
    rrr         = make_module_reload_func(None, module_prefix, module)
    profile_    = make_module_profile_func(None, module_prefix, module)
    print_funcs = inject_print_functions(None, module_prefix, DEBUG, module)
    (print, print_, printDBG) = print_funcs
    return (print, print_, printDBG, rrr, profile_)


def inject2(module_name=None, module_prefix=None, DEBUG=False, module=None, N=1):
    """ wrapper that depricates print_ and printDBG """
    if module_prefix is None:
        module_prefix = '[%s]' % (module_name,)
    noinject(module_name, module_prefix, DEBUG, module, N=N)
    module = _get_module(module_name, module)
    rrr      = make_module_reload_func(None, module_prefix, module)
    profile_ = make_module_profile_func(None, module_prefix, module)
    print    = make_module_print_func(module)
    return print, rrr, profile_


def split_python_text_into_lines(text):
    """
    # TODO: make it so this function returns text so one statment is on one
    # line that means no splitting up things like function definitions into
    # multiple lines
    """
    #import jedi
    #script = jedi.Script(text, line=1, column=None, path='')
    def parentesis_are_balanced(line):
        """
        helper

        References:
            http://stackoverflow.com/questions/18007995/recursive-paren-balance
        """
        def balanced(str_, i=0, cnt=0, left='(', right=')'):
            if i == len(str_):
                return cnt == 0
            if cnt < 0:
                return False
            if str_[i] == left:
                return  balanced(str_, i + 1, cnt + 1)
            elif str_[i] == right:
                return  balanced(str_, i + 1, cnt - 1)
            return balanced(str_, i + 1, cnt)
        return balanced(line)

    lines = text.split('\n')
    new_lines = []
    current_line = ''
    for line in lines:
        current_line += line
        if parentesis_are_balanced(current_line):
            new_lines.append(current_line)
            current_line = ''
    return lines


def inject_python_code2(fpath, patch_code, tag):
    """ Does autogeneration stuff """
    import utool as ut
    text = ut.readfrom(fpath)
    start_tag = '# <%s>' % tag
    end_tag = '# </%s>' % tag
    new_text = ut.replace_between_tags(text, patch_code, start_tag, end_tag)
    ut.writeto(fpath, new_text)


def inject_python_code(fpath, patch_code, tag=None,
                       inject_location='after_imports'):
    """
    DEPRICATE
    puts code into files on disk
    """
    import utool as ut
    assert tag is not None, 'TAG MUST BE SPECIFIED IN INJECTED CODETEXT'
    text = ut.read_from(fpath)
    comment_start_tag = '# <util_inject:%s>' % tag
    comment_end_tag  = '# </util_inject:%s>' % tag

    tagstart_txtpos = text.find(comment_start_tag)
    tagend_txtpos = text.find(comment_end_tag)

    text_lines = ut.split_python_text_into_lines(text)

    # split the file into two parts and inject code between them
    if tagstart_txtpos != -1 or tagend_txtpos != -1:
        assert tagstart_txtpos != -1, 'both tags must not be found'
        assert tagend_txtpos != -1, 'both tags must not be found'

        for pos, line in enumerate(text_lines):
            if line.startswith(comment_start_tag):
                tagstart_pos = pos
            if line.startswith(comment_end_tag):
                tagend_pos = pos
        part1 = text_lines[0:tagstart_pos]
        part2 = text_lines[tagend_pos + 1:]
    else:
        if inject_location == 'after_imports':
            first_nonimport_pos = 0
            for line in text_lines:
                list_ = ['import ', 'from ', '#', ' ']
                isvalid = (len(line) == 0 or
                           any(line.startswith(str_) for str_ in list_))
                if not isvalid:
                    break
                first_nonimport_pos += 1
            part1 = text_lines[0:first_nonimport_pos]
            part2 = text_lines[first_nonimport_pos:]
        else:
            raise AssertionError('Unknown inject location')

    newtext = (
        '\n'.join(part1 + [comment_start_tag]) +
        '\n' + patch_code + '\n' +
        '\n'.join( [comment_end_tag] + part2)
    )
    text_backup_fname = fpath + '.' + ut.get_timestamp() + '.bak'
    ut.write_to(text_backup_fname, text)
    ut.write_to(fpath, newtext)
    #print(newtext)


if '--inject-color' in sys.argv or '--cex' in sys.argv:
    inject_colored_exceptions()

# Inject this module with itself!
print, rrr, profile = inject2(__name__, '[inject]')


if __name__ == '__main__':
    """
    CommandLine:
        python -m utool.util_inject
        python -m utool.util_inject --allexamples
        python -m utool.util_inject --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
import utool as ut


if __name__ == '__main__':
    """
    SeeAlso:
        utool.util_autogen.makeinit
    """
    ut.execute_doctest(ut.makeinit, testnum=0)



# -*- coding: utf-8 -*-
"""
Ignore:
    ~/local/init/REPOS1.py
"""
from __future__ import absolute_import, division, print_function  # , unicode_literals
from six.moves import zip, filter, filterfalse, map, range  # NOQA
import six  # NOQA
#from os.path import split, dirname, join
from os.path import dirname, join
from utool import util_class  # NOQA
from utool import util_dev
from utool import util_inject
print, rrr, profile = util_inject.inject2(__name__)


__GLOBAL_PROFILE__ = None


def ensure_text(fname, text, repo_dpath='.', force=None, locals_={}, chmod=None):
    """
    Args:
        fname (str):  file name
        text (str):
        repo_dpath (str):  directory path string(default = '.')
        force (bool): (default = False)
        locals_ (dict): (default = {})

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_project import *  # NOQA
        >>> import utool as ut
        >>> result = setup_repo()
        >>> print(result)
    """
    import utool as ut
    ut.colorprint('Ensuring fname=%r' % (fname), 'yellow')

    # if not fname.endswith('__init__.py'):
    #     # HACK
    #     return

    if force is None and ut.get_argflag('--force-%s' % (fname,)):
        force = True
    text_ = ut.remove_codeblock_syntax_sentinals(text)
    fmtkw = locals_.copy()
    fmtkw['fname'] = fname
    text_ = text_.format(**fmtkw) + '\n'

    fpath = join(repo_dpath, fname)
    ut.dump_autogen_code(fpath, text_)

    # if force or not ut.checkpath(fpath, verbose=2, n=5):
    #     ut.writeto(fpath, text_)
    #     try:
    #         if chmod:
    #             ut.chmod(fpath, chmod)
    #     except Exception as ex:
    #         ut.printex(ex, iswarning=True)
    # else:
    #     print(ut.color_diff_text(ut.difftext(ut.readfrom(fpath), text_)))
    #     print('use -w to force write')


class SetupRepo(object):
    """
    Maybe make a new interface to SetupRepo?

    Example:
        >>> # DISABLE_DOCTEST
        >>> # SCRIPT
        >>> from utool.util_project import *  # NOQA
        >>> import utool as ut
        >>> self = SetupRepo()
        >>> print(self)
    """

    def __init__(self):
        import utool as ut
        self.modname = None
        code_dpath = ut.truepath(ut.get_argval('--code-dir', default='~/code'))
        self.code_dpath = ut.unexpanduser(code_dpath)
        self.repo_fname = (ut.get_argval(('--repo', '--repo-name'), type_=str))
        self.repo_dpath = join(code_dpath, self.repo_fname)
        self.modname = ut.get_argval('--modname', default=self.repo_fname)
        self.regenfmt = 'python -m utool SetupRepo.{cmd} --modname={modname} --repo={repo_fname} --codedir={code_dpath}'
        ut.ensuredir(self.repo_dpath, verbose=True)

    def all(self):
        pass

    def ensure_text(self, fname, text, **kwargs):
        ensure_text(fname, text, locals_=self.__dict__,
                    repo_dpath=self.repo_dpath, **kwargs)

    def main(self):
        """
        python -m utool SetupRepo.main --modname=sklearn --repo=scikit-learn --codedir=~/code -w
        python -m utool SetupRepo.main --repo=ubelt --codedir=~/code --modname=ubelt -w

        Example:
        >>> # DISABLE_DOCTEST
            >>> # SCRIPT
            >>> from utool.util_project import *  # NOQA
            >>> SetupRepo().main()
        """
        self.regencmd = self.regenfmt.format(cmd='main', **self.__dict__)
        import utool as ut
        self.ensure_text(
            fname=join(self.modname, '__main__.py'),
            chmod='+x',
            text=ut.codeblock(
                r'''
                # STARTBLOCK
                #!/usr/bin/env python
                # -*- coding: utf-8 -*-
                """
                Initially Generated By:
                    {regencmd}
                """
                from __future__ import absolute_import, division, print_function, unicode_literals


                def {modname}_main():
                    ignore_prefix = []
                    ignore_suffix = []
                    import utool as ut
                    ut.main_function_tester('{modname}', ignore_prefix, ignore_suffix)

                if __name__ == '__main__':
                    """
                    Usage:
                        python -m {modname} <funcname>
                    """
                    print('Running {modname} main')
                    {modname}_main()
                # ENDBLOCK
                '''
            )
        )


def setup_repo():
    r"""
    Creates default structure for a new repo

    CommandLine:
        python -m utool setup_repo --repo=dtool --codedir=~/code

        python -m utool setup_repo --repo=dtool --codedir=~/code
        python -m utool setup_repo --repo=ibeis-flukematch-module --codedir=~/code --modname=ibeis_flukematch
        python -m utool setup_repo --repo=mtgmonte --codedir=~/code --modname=mtgmonte
        python -m utool setup_repo --repo=pydarknet --codedir=~/code --modname=pydarknet
        python -m utool setup_repo --repo=sandbox_utools --codedir=~/code --modname=sandbox_utools

        python -m utool setup_repo --repo=ubelt --codedir=~/code --modname=ubelt -w

        python -m utool setup_repo

    Python:
        ipython
        import utool as ut
        ut.rrrr(0); ut.setup_repo()

    Example:
        >>> # DISABLE_DOCTEST
        >>> # SCRIPT
        >>> from utool.util_project import *  # NOQA
        >>> import utool as ut
        >>> result = setup_repo()
        >>> print(result)
    """
    print('\n [setup_repo]!')
    # import os
    from functools import partial
    import utool as ut
    # import os
    code_dpath  = ut.truepath(ut.get_argval('--code-dir', default='~/code'))
    _code_dpath = ut.unexpanduser(code_dpath)
    repo_fname = (ut.get_argval(('--repo', '--repo-name'), type_=str))
    repo_dpath = join(code_dpath, repo_fname)
    modname = ut.get_argval('--modname', default=repo_fname)
    ut.ensuredir(repo_dpath, verbose=True)
    _regencmd = 'python -m utool --tf setup_repo --repo={repo_fname} --codedir={_code_dpath} --modname={modname}'
    flake8_noqacmd = 'flake8' + ':noqa'
    regencmd = _regencmd.format(**locals())
    with ut.ChdirContext(repo_dpath):
        # os.chdir(repo_fname)
        locals_ = locals()
        force = True

        _ensure_text = partial(ensure_text, repo_dpath='.', force=None, locals_=locals_)

        _ensure_text(
            fname='todo.md',
            text=ut.codeblock(
                r'''
                # STARTBLOCK
                # {modname} TODO File

                * Add TODOS!
                # ENDBLOCK
                ''')
        )

        _ensure_text(
            fname='README.md',
            text=ut.codeblock(
                r'''
                # STARTBLOCK
                # {modname} README FILE
                # ENDBLOCK
                ''')
        )

        _ensure_text(
            fname='setup.py',
            chmod='+x',
            text=ut.codeblock(
                r'''
                # STARTBLOCK
                #!/usr/bin/env python
                """
                Initially Generated By:
                    {regencmd} --force-{fname}
                """
                from __future__ import absolute_import, division, print_function, unicode_literals
                from setuptools import setup
                try:
                    from utool import util_setup
                except ImportError:
                    print('ERROR: setup requires utool')
                    raise

                INSTALL_REQUIRES = [
                    #'cython >= 0.21.1',
                    #'numpy >= 1.9.0',
                    #'scipy >= 0.16.0',
                ]

                CLUTTER_PATTERNS = [
                    # Patterns removed by python setup.py clean
                ]

                if __name__ == '__main__':
                    kwargs = util_setup.setuptools_setup(
                        setup_fpath=__file__,
                        name='{modname}',
                        packages=util_setup.find_packages(),
                        version=util_setup.parse_package_for_version('{modname}'),
                        license=util_setup.read_license('LICENSE'),
                        long_description=util_setup.parse_readme('README.md'),
                        ext_modules=util_setup.find_ext_modules(),
                        cmdclass=util_setup.get_cmdclass(),
                        #description='description of module',
                        #url='https://github.com/<username>/{repo_fname}.git',
                        #author='<author>',
                        #author_email='<author_email>',
                        keywords='',
                        install_requires=INSTALL_REQUIRES,
                        clutter_patterns=CLUTTER_PATTERNS,
                        #package_data={{'build': ut.get_dynamic_lib_globstrs()}},
                        #build_command=lambda: ut.std_build_command(dirname(__file__)),
                        classifiers=[],
                    )
                    setup(**kwargs)
                # ENDBLOCK
                '''
            )
        )

        _ensure_text(
            fname='.gitignore',
            text=ut.codeblock(
                r'''
                # STARTBLOCK
                *.py[cod]

                # C extensions
                *.so
                # Packages
                *.egg
                *.egg-info
                dist
                build
                eggs
                parts
                bin
                var
                sdist
                develop-eggs
                .installed.cfg
                lib
                lib64
                __pycache__

                # Installer logs
                pip-log.txt

                # Print Logs
                logs

                # Unit test / coverage reports
                .coverage
                .tox
                nosetests.xml

                # Translations
                *.mo

                # Mr Developer
                .mr.developer.cfg
                .project
                .pydevproject
                .DS_Store
                *.dump.txt
                *.sqlite3

                # profiler
                *.lprof
                *.prof

                *.flann
                *.npz

                # utool output
                _timeings.txt
                failed.txt

                *.orig
                _doc
                timeings.txt
                failed_doctests.txt
                # ENDBLOCK
                '''
            )
        )

        _ensure_text(
            fname=join(repo_dpath, modname, '__init__.py'),
            text=ut.codeblock(
                r'''
                # STARTBLOCK
                # -*- coding: utf-8 -*-
                # {flake8_noqacmd}
                """
                Initially Generated By:
                    {regencmd}
                """
                from __future__ import absolute_import, division, print_function, unicode_literals
                import sys
                __version__ = '0.0.0'

                IMPORT_TUPLES = [
                    # ('<modname>', None),
                ]
                __DYNAMIC__ = '--nodyn' not in sys.argv

                """
                python -c "import {modname}" --dump-{modname}-init
                python -c "import {modname}" --update-{modname}-init
                """

                DOELSE = False
                if __DYNAMIC__:
                    # Dynamically import listed util libraries and their members.
                    from utool._internal import util_importer
                    ignore_endswith = []
                    import_execstr = util_importer.dynamic_import(
                        __name__, IMPORT_TUPLES, ignore_endswith=ignore_endswith)
                    exec(import_execstr)
                    DOELSE = False
                else:
                    DOELSE = True
                if DOELSE:
                    # <AUTOGEN_INIT>
                    pass
                    # </AUTOGEN_INIT>
                # ENDBLOCK
                '''
            )
        )

        _ensure_text(
            fname=join(repo_dpath, modname, '__main__.py'),
            chmod='+x',
            text=ut.codeblock(
                r'''
                # STARTBLOCK
                #!/usr/bin/env python
                # -*- coding: utf-8 -*-
                """
                Initially Generated By:
                    {regencmd}
                """
                from __future__ import absolute_import, division, print_function, unicode_literals


                def {modname}_main():
                    ignore_prefix = []
                    ignore_suffix = []
                    import utool as ut
                    ut.main_function_tester('{modname}', ignore_prefix, ignore_suffix)

                if __name__ == '__main__':
                    """
                    Usage:
                        python -m {modname} <funcname>
                    """
                    print('Running {modname} main')
                    {modname}_main()
                # ENDBLOCK
                '''
            )
        )

        _ensure_text(
            fname='run_tests.py',
            chmod='+x',
            text=ut.codeblock(
                r'''
                # STARTBLOCK
                #!/usr/bin/env python
                """
                Initially Generated By:
                    {regencmd} --force-{fname}
                """
                from __future__ import absolute_import, division, print_function
                import sys
                import utool as ut


                def run_tests():
                    # Build module list and run tests
                    import sys
                    ut.change_term_title('RUN {modname} TESTS')
                    exclude_doctests_fnames = set([
                    ])
                    exclude_dirs = [
                        '_broken', 'old', 'tests', 'timeits',
                        '_scripts', '_timeits', '_doc', 'notebook',
                    ]
                    dpath_list = ['{modname}']
                    doctest_modname_list = ut.find_doctestable_modnames(
                        dpath_list, exclude_doctests_fnames, exclude_dirs)

                    coverage = ut.get_argflag(('--coverage', '--cov',))
                    if coverage:
                        import coverage
                        cov = coverage.Coverage(source=doctest_modname_list)
                        cov.start()
                        print('Starting coverage')

                        exclude_lines = [
                            'pragma: no cover',
                            'def __repr__',
                            'if self.debug:',
                            'if settings.DEBUG',
                            'raise AssertionError',
                            'raise NotImplementedError',
                            'if 0:',
                            'if ut.VERBOSE',
                            'if _debug:',
                            'if __name__ == .__main__.:',
                            'print(.*)',
                        ]
                        for line in exclude_lines:
                            cov.exclude(line)

                    for modname in doctest_modname_list:
                        exec('import ' + modname, globals())
                    module_list = [sys.modules[name] for name in doctest_modname_list]

                    nPass, nTotal, failed_cmd_list = ut.doctest_module_list(module_list)

                    if coverage:
                        print('Stoping coverage')
                        cov.stop()
                        print('Saving coverage')
                        cov.save()
                        print('Generating coverage html report')
                        cov.html_report()

                    if nPass != nTotal:
                        return 1
                    else:
                        return 0

                if __name__ == '__main__':
                    import multiprocessing
                    multiprocessing.freeze_support()
                    retcode = run_tests()
                    sys.exit(retcode)
                # ENDBLOCK
                '''
            )
        )

    ut.ensuredir(join(repo_dpath, modname), verbose=True)


#@util_class.ReloadingMetaclass
class UserProfile(util_dev.NiceRepr):
    def __nice__(self):
        num_repos = 0 if self.project_dpaths is None else len(self.project_dpaths)
        return str(self.project_name) + ' %d repos' % (num_repos,)

    def __init__(self, name=None):
        self.project_name = name
        self.project_dpaths = None
        self.project_include_patterns = None
        self.project_exclude_dirs = []
        self.project_exclude_patterns = []

    def grep(self, *args, **kwargs):
        return grep_projects(user_profile=self, *args, **kwargs)

    def glob(self, *args, **kwargs):
        r"""
        # Ensure that .gitignore has certain lines
        git_ignore_lines = [
            'timeings.txt'
        ]
        fpath_list = profile.glob('.gitignore', recursive=False)
        for fpath in fpath_list:
            lines = ut.readfrom(fpath, verbose=False).split('\n')
            lines = [line.strip() for line in lines]
            missing = ut.setdiff(git_ignore_lines, lines)
            if missing:
                print('fpath = %r' % (fpath,))
                ut.writeto(fpath, '\n'.join(lines + missing))
        """
        return glob_projects(user_profile=self, *args, **kwargs)

    # def __str__(self):
    #     return


def ibeis_user_profile():
    import utool as ut
    import sys
    addpath = True
    module_fpath = ut.truepath('~/local/init/REPOS1.py')
    if addpath:
        module_dpath = dirname(module_fpath)
        sys.path.append(module_dpath)
    REPOS1 = ut.import_module_from_fpath(module_fpath)
    self = UserProfile(name='ibeis')
    #self.project_dpaths = REPOS1.PROJECT_REPOS
    self.project_dpaths = REPOS1.IBEIS_REPOS
    # self.project_dpaths += [ut.truepath('~/latex/crall-candidacy-2015/')]
    self.project_dpaths += [
        ut.truepath('~/local'),
        ut.truepath('~/code/fixtex'),
        ut.truepath('~/code/pyrf'),
        ut.truepath('~/code/detecttools'),
        ut.truepath('~/code/pydarknet'),
    ]
    self.project_dpaths = ut.unique(self.project_dpaths)
    # self.project_dpaths += [ut.truepath('~/local/vim/rc')]
    self.project_include_patterns = [
        '*.py', '*.cxx', '*.cpp', '*.hxx', '*.hpp', '*.c', '*.h', '*.vim'
        #'*.py',  # '*.cxx', '*.cpp', '*.hxx', '*.hpp', '*.c', '*.h', '*.vim'
    ]
    self.project_exclude_dirs = [
        '_graveyard', '_broken', 'CompilerIdCXX', 'CompilerIdC', 'build',
        'old', '_old_qt_hs_matcher', 'htmlcov'
    ]
    self.project_exclude_patterns = ['_grave*', '_autogen_explicit_controller*']
    return self


def ensure_user_profile(user_profile=None):
    r"""
    Args:
        user_profile (UserProfile): (default = None)

    Returns:
        UserProfile: user_profile

    CommandLine:
        python -m utool.util_project --exec-ensure_user_profile --show

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_project import *  # NOQA
        >>> import utool as ut
        >>> user_profile = None
        >>> user_profile = ensure_user_profile(user_profile)
        >>> result = ('user_profile = %s' % (ut.repr2(user_profile),))
        >>> print(ut.repr3(user_profile.project_dpaths))
        >>> print(result)
    """
    global __GLOBAL_PROFILE__
    if __GLOBAL_PROFILE__ is None:
        import utool as ut
        if ut.is_developer():
            __GLOBAL_PROFILE__ = ibeis_user_profile()
        else:
            __GLOBAL_PROFILE__ = UserProfile('default')
    if user_profile is None:
        user_profile = __GLOBAL_PROFILE__
    return user_profile


def grep_projects(tofind_list, user_profile=None, verbose=True, new=False,
                  **kwargs):
    r"""
    Greps the projects defined in the current UserProfile

    Args:
        tofind_list (list):
        user_profile (None): (default = None)

    Kwargs:
        user_profile

    CommandLine:
        python -m utool --tf grep_projects grep_projects

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_project import *  # NOQA
        >>> import utool as ut
        >>> import sys
        >>> tofind_list = ut.get_argval('--find', type_=list,
        >>>                             default=[sys.argv[-1]])
        >>> grep_projects(tofind_list)
    """
    import utool as ut
    user_profile = ensure_user_profile(user_profile)
    print('user_profile = {!r}'.format(user_profile))

    kwargs = kwargs.copy()
    colored = kwargs.pop('colored', True)

    grepkw = {}
    grepkw['greater_exclude_dirs'] = user_profile.project_exclude_dirs
    grepkw['exclude_dirs'] = user_profile.project_exclude_dirs
    grepkw['dpath_list'] = user_profile.project_dpaths
    grepkw['include_patterns'] = user_profile.project_include_patterns
    grepkw['exclude_patterns'] = user_profile.project_exclude_patterns
    grepkw.update(kwargs)

    msg_list1 = []
    msg_list2 = []

    print_ = msg_list1.append
    print_('Greping Projects')
    print_('tofind_list = %s' % (ut.repr4(tofind_list, nl=True),))
    #print_('grepkw = %s' % ut.repr4(grepkw, nl=True))
    if verbose:
        print('\n'.join(msg_list1))
    #with ut.Timer('greping', verbose=True):
    grep_result = ut.grep(tofind_list, **grepkw)
    found_fpath_list, found_lines_list, found_lxs_list = grep_result

    # HACK, duplicate behavior. TODO: write grep print result function
    reflags = grepkw.get('reflags', 0)
    _exprs_flags = [ut.extend_regex2(expr, reflags)
                    for expr in tofind_list]
    extended_regex_list = ut.take_column(_exprs_flags, 0)
    reflags_list = ut.take_column(_exprs_flags, 1)
    # HACK
    # pat = ut.util_regex.regex_or(extended_regex_list)
    reflags = reflags_list[0]

    # from utool import util_regex
    resultstr = ut.make_grep_resultstr(grep_result, extended_regex_list,
                                       reflags, colored=colored)
    msg_list2.append(resultstr)
    print_ = msg_list2.append
    #for fpath, lines, lxs in zip(found_fpath_list, found_lines_list,
    #                             found_lxs_list):
    #    print_('----------------------')
    #    print_('found %d line(s) in %r: ' % (len(lines), fpath))
    #    name = split(fpath)[1]
    #    max_line = len(lines)
    #    ndigits = str(len(str(max_line)))
    #    for (lx, line) in zip(lxs, lines):
    #        line = line.replace('\n', '')
    #        print_(('%s : %' + ndigits + 'd |%s') % (name, lx, line))
    # iter_ = zip(found_fpath_list, found_lines_list, found_lxs_list)
    # for fpath, lines, lxs in iter_:
    #     print_('----------------------')
    #     print_('found %d line(s) in %r: ' % (len(lines), fpath))
    #     name = split(fpath)[1]
    #     max_line = len(lines)
    #     ndigits = str(len(str(max_line)))
    #     for (lx, line) in zip(lxs, lines):
    #         line = line.replace('\n', '')
    #         colored_line = ut.highlight_regex(
    #             line.rstrip('\n'), pat, reflags=reflags)
    #         print_(('%s : %' + ndigits + 'd |%s') % (name, lx, colored_line))

    print_('====================')
    print_('found_fpath_list = ' + ut.repr4(found_fpath_list))
    print_('')
    #print_('gvim -o ' + ' '.join(found_fpath_list))
    if verbose:
        print('\n'.join(msg_list2))
    msg_list = msg_list1 + msg_list2

    if new:
        return GrepResult(found_fpath_list, found_lines_list, found_lxs_list,
                          extended_regex_list, reflags)
    else:
        return msg_list


def glob_projects(pat, user_profile=None, recursive=True):
    """

    def testenv(modname, funcname):
        ut.import_modname(modname)
        exec(ut.execstr_funckw(table.get_rowid), globals())

    Ignore:
        >>> import utool as ut
        >>> ut.testenv('utool.util_project', 'glob_projects', globals())
        >>> from utool.util_project import *  # NOQA
    """
    import utool as ut  # NOQA
    user_profile = ensure_user_profile(user_profile)
    glob_results = ut.flatten([ut.glob(dpath, pat, recursive=recursive,
                                       exclude_dirs=user_profile.project_exclude_dirs)
                               for dpath in user_profile.project_dpaths])
    return glob_results


class GrepResult(util_dev.NiceRepr):
    def __init__(self, found_fpath_list, found_lines_list,
                 found_lxs_list, extended_regex_list, reflags):
        self.found_fpath_list = found_fpath_list
        self.found_lines_list = found_lines_list
        self.found_lxs_list = found_lxs_list
        self.extended_regex_list = extended_regex_list
        self.reflags = reflags
        self.filter_pats = []

    def __nice__(self):
        return '(%d)' % (len(self),)

    def __str__(self):
        return self.make_resultstr()

    def __len__(self):
        return len(self.found_fpath_list)

    def __delitem__(self, index):
        import utool as ut
        index = ut.ensure_iterable(index)
        ut.delete_items_by_index(self.found_fpath_list, index)
        ut.delete_items_by_index(self.found_lines_list, index)
        ut.delete_items_by_index(self.found_lxs_list, index)

    def __getitem__(self, index):
        return (
            ut.take(self.found_fpath_list, index),
            ut.take(self.found_lines_list, index),
            ut.take(self.found_lxs_list, index),
        )

    def remove_results(self, indicies):
        del self[indicies]

    def make_resultstr(self, colored=True):
        import utool as ut
        tup = (self.found_fpath_list, self.found_lines_list,
               self.found_lxs_list)
        return ut.make_grep_resultstr(tup, self.extended_regex_list,
                                      self.reflags, colored=colored)

    # def make_big_resultstr():
    #     pass

    def pattern_filterflags(self, filter_pat):
        self.filter_pats.append(filter_pat)
        import re
        flags_list = [[re.search(filter_pat, line) is None for line in lines]
                      for fpath, lines, lxs in zip(self.found_fpath_list, self.found_lines_list, self.found_lxs_list)]
        return flags_list

    def inplace_filter_results(self, filter_pat):
        import utool as ut
        self.filter_pats.append(filter_pat)
        # Get zipflags
        flags_list = self.pattern_filterflags(filter_pat)
        # Check to see if there are any survivors
        flags = ut.lmap(any, flags_list)
        #
        found_lines_list = ut.zipcompress(self.found_lines_list, flags_list)
        found_lxs_list = ut.zipcompress(self.found_lxs_list, flags_list)
        #
        found_fpath_list = ut.compress(self.found_fpath_list, flags)
        found_lines_list = ut.compress(found_lines_list, flags)
        found_lxs_list = ut.compress(found_lxs_list, flags)
        # In place modification
        self.found_fpath_list = found_fpath_list
        self.found_lines_list = found_lines_list
        self.found_lxs_list = found_lxs_list

    def hack_remove_pystuff(self):
        import utool as ut
        # Hack of a method
        new_lines = []
        for lines in self.found_lines_list:
            # remove comment results
            flags = [not line.strip().startswith('# ') for line in lines]
            lines = ut.compress(lines, flags)

            # remove doctest results
            flags = [not line.strip().startswith('>>> ') for line in lines]
            lines = ut.compress(lines, flags)

            # remove cmdline tests
            import re
            flags = [not re.search('--test-' + self.extended_regex_list[0], line) for line in lines]
            lines = ut.compress(lines, flags)

            flags = [not re.search('--exec-' + self.extended_regex_list[0], line) for line in lines]
            lines = ut.compress(lines, flags)

            flags = [not re.search('--exec-[a-zA-z]*\\.' + self.extended_regex_list[0], line) for line in lines]
            lines = ut.compress(lines, flags)

            flags = [not re.search('--test-[a-zA-z]*\\.' + self.extended_regex_list[0], line) for line in lines]
            lines = ut.compress(lines, flags)

            # remove func defs
            flags = [not re.search('def ' + self.extended_regex_list[0], line) for line in lines]
            lines = ut.compress(lines, flags)
            new_lines += [lines]
        self.found_lines_list = new_lines

        # compress self
        flags = [len(lines_) > 0 for lines_ in self.found_lines_list]
        idxs = ut.list_where(ut.not_list(flags))
        del self[idxs]


## Grep my projects
#def gp(r, regexp):
#    rob_nav._grep(r, [regexp], recursive=True, dpath_list=project_dpaths(), regex=True)

## Sed my projects
#def sp(r, regexpr, repl, force=False):
#    rob_nav._sed(r, regexpr, repl, force=force, recursive=True, dpath_list=project_dpaths())

def sed_projects(regexpr, repl, force=False, recursive=True, user_profile=None, **kwargs):
    r"""

    Args:
        regexpr (?):
        repl (?):
        force (bool): (default = False)
        recursive (bool): (default = True)
        user_profile (None): (default = None)

    CommandLine:
        python -m utool.util_project --exec-sed_projects

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_project import *  # NOQA
        >>> regexpr = ut.get_argval('--find', type_=str, default=sys.argv[-1])
        >>> repl = ut.get_argval('--repl', type_=str, default=sys.argv[-2])
        >>> force = False
        >>> recursive = True
        >>> user_profile = None
        >>> result = sed_projects(regexpr, repl, force, recursive, user_profile)
        >>> print(result)

    Ignore:
        regexpr = 'annotation match_scores'
        repl = 'draw_annot_scoresep'

    """
    # FIXME: finishme
    import utool as ut
    user_profile = ensure_user_profile(user_profile)

    sedkw = {}
    sedkw['exclude_dirs'] = user_profile.project_exclude_dirs
    sedkw['dpath_list'] = user_profile.project_dpaths
    sedkw['include_patterns'] = user_profile.project_include_patterns
    sedkw.update(kwargs)

    msg_list1 = []
    #msg_list2 = []

    print_ = msg_list1.append
    print_('Seding Projects')
    print(' * regular expression : %r' % (regexpr,))
    print(' * replacement        : %r' % (repl,))
    print_('sedkw = %s' % ut.repr4(sedkw, nl=True))

    print(' * recursive: %r' % (recursive,))
    print(' * force: %r' % (force,))

    # Walk through each directory recursively
    for fpath in ut.matching_fpaths(sedkw['dpath_list'],
                                    sedkw['include_patterns'],
                                    sedkw['exclude_dirs'],
                                    recursive=recursive):
        ut.sedfile(fpath, regexpr, repl, force)


#def extend_regex(regexpr):
#    regex_map = {
#        r'\<': r'\b(?=\w)',
#        r'\>': r'\b(?!\w)',
#        ('UNSAFE', r'\x08'): r'\b',
#    }
#    for key, repl in six.iteritems(regex_map):
#        if isinstance(key, tuple):
#            search = key[1]
#        else:
#            search = key
#        if regexpr.find(search) != -1:
#            if isinstance(key, tuple):
#                print('WARNING! Unsafe regex with: %r' % (key,))
#            regexpr = regexpr.replace(search, repl)
#    return regexpr
#regexpr = extend_regex(regexpr)
#if '\x08' in regexpr:
#    print('Remember \\x08 != \\b')
#    print('subsituting for you for you')
#    regexpr = regexpr.replace('\x08', '\\b')
#    print(' * regular expression : %r' % (regexpr,))


if False:
    def ensure_vim_plugins():
        """

        python ~/local/init/ensure_vim_plugins.py
        '~/local/init/ensure_vim_plugins.py'
        '~/local/init/REPOS1.py'

        """
        # TODO
        pass

    def find_module_callers():
        """
        TODO:
        attempt to build a call graph between module functions to make it easy to see
        what can be removed and what cannot.
        """
        import utool as ut
        from os.path import normpath
        mod_fpath = ut.truepath('~/code/ibeis/ibeis/expt/results_analyzer.py')
        mod_fpath = ut.truepath('~/code/ibeis/ibeis/expt/results_all.py')
        mod_fpath = ut.truepath('~/code/ibeis/ibeis/expt/results_organizer.py')
        module = ut.import_module_from_fpath(mod_fpath)
        user_profile = ut.ensure_user_profile()
        doctestables = list(ut.iter_module_doctestable(module, include_builtin=False))
        grepkw = {}
        grepkw['exclude_dirs'] = user_profile.project_exclude_dirs
        grepkw['dpath_list'] = user_profile.project_dpaths
        grepkw['verbose'] = True

        usage_map = {}
        for funcname, func in doctestables:
            print('Searching for funcname = %r' % (funcname,))
            found_fpath_list, found_lines_list, found_lxs_list = ut.grep([funcname], **grepkw)
            used_in = (found_fpath_list, found_lines_list, found_lxs_list)
            usage_map[funcname] = used_in

        external_usage_map = {}
        for funcname, used_in in usage_map.items():
            (found_fpath_list, found_lines_list, found_lxs_list) = used_in
            isexternal_flag = [normpath(fpath) != normpath(mod_fpath) for fpath in found_fpath_list]
            ext_used_in = (ut.compress(found_fpath_list, isexternal_flag),
                           ut.compress(found_lines_list, isexternal_flag),
                           ut.compress(found_lxs_list, isexternal_flag))
            external_usage_map[funcname] = ext_used_in

        for funcname, used_in in external_usage_map.items():
            (found_fpath_list, found_lines_list, found_lxs_list) = used_in

        print('Calling modules: \n' +
              ut.repr2(ut.unique_ordered(ut.flatten([used_in[0] for used_in in  external_usage_map.values()])), nl=True))


if __name__ == '__main__':
    """
    CommandLine:
        python -m utool.util_project
        python -m utool.util_project --allexamples
        python -m utool.util_project --allexamples --noface --nosrc
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function
import six
import itertools
import random
try:
    import numpy as np
except ImportError:
    pass
from utool import util_inject
print, rrr, profile = util_inject.inject2(__name__)


def tiled_range(range_, cols):
    return np.tile(np.arange(range_), (cols, 1)).T
    #np.tile(np.arange(num_qf).reshape(num_qf, 1), (1, k_vsmany))


def quantum_random():
    """ returns a 32 bit unsigned integer quantum random number """
    import quantumrandom
    data16 = quantumrandom.uint16(array_length=2)
    assert data16.flags['C_CONTIGUOUS']
    data32 = data16.view(np.dtype('uint32'))[0]
    return data32


def _npstate_to_pystate(npstate):
    """
    Convert state of a NumPy RandomState object to a state
    that can be used by Python's Random.

    References:
        https://stackoverflow.com/questions/44313620/converting-randomstate

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_numpy import *  # NOQA
        >>> from utool.util_numpy import _npstate_to_pystate
        >>> py_rng = random.Random(0)
        >>> np_rng = np.random.RandomState(seed=0)
        >>> npstate = np_rng.get_state()
        >>> pystate = _npstate_to_pystate(npstate)
        >>> py_rng.setstate(pystate)
        >>> assert np_rng.rand() == py_rng.random()
    """
    PY_VERSION = 3
    version, keys, pos, has_gauss, cached_gaussian_ = npstate
    keys_pos = tuple(map(int, keys)) + (int(pos),)
    cached_gaussian_ = cached_gaussian_ if has_gauss else None
    pystate = (PY_VERSION, keys_pos, cached_gaussian_)
    return pystate


def _pystate_to_npstate(pystate):
    """
    Convert state of a Python Random object to state usable
    by NumPy RandomState.

    References:
        https://stackoverflow.com/questions/44313620/converting-randomstate

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_numpy import *  # NOQA
        >>> from utool.util_numpy import _pystate_to_npstate
        >>> py_rng = random.Random(0)
        >>> np_rng = np.random.RandomState(seed=0)
        >>> pystate = py_rng.getstate()
        >>> npstate = _pystate_to_npstate(pystate)
        >>> np_rng.set_state(npstate)
        >>> assert np_rng.rand() == py_rng.random()
    """
    NP_VERSION = 'MT19937'
    version, keys_pos_, cached_gaussian_ = pystate
    keys, pos = keys_pos_[:-1], keys_pos_[-1]
    keys = np.array(keys, dtype=np.uint32)
    has_gauss = cached_gaussian_ is not None
    cached_gaussian = cached_gaussian_ if has_gauss else 0.0
    npstate = (NP_VERSION, keys, pos, has_gauss, cached_gaussian)
    return npstate


def ensure_rng(rng, impl='numpy'):
    """
    Returns a random number generator

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_numpy import *  # NOQA
        >>> import utool as ut
        >>> import numpy as np
        >>> num = 4
        >>> print('--- Python as PYTHON ---')
        >>> py_rng = random.Random(0)
        >>> pp_nums = [py_rng.random() for _ in range(num)]
        >>> print(pp_nums)
        >>> print('--- Numpy as PYTHON ---')
        >>> np_rng = ut.ensure_rng(random.Random(0), impl='numpy')
        >>> np_nums = [np_rng.rand() for _ in range(num)]
        >>> print(np_nums)
        >>> print('--- Numpy as NUMPY---')
        >>> np_rng = np.random.RandomState(seed=0)
        >>> nn_nums = [np_rng.rand() for _ in range(num)]
        >>> print(nn_nums)
        >>> print('--- Python as NUMPY---')
        >>> py_rng = ut.ensure_rng(np.random.RandomState(seed=0), impl='python')
        >>> pn_nums = [py_rng.random() for _ in range(num)]
        >>> print(pn_nums)
        >>> assert np_nums == pp_nums
        >>> assert pn_nums == nn_nums
    """
    if impl == 'numpy':
        if rng is None:
            rng = np.random
        elif isinstance(rng, int):
            rng = np.random.RandomState(seed=rng)
        elif isinstance(rng, random.Random):
            # Convert python to numpy random state
            py_rng = rng
            pystate = py_rng.getstate()
            npstate = _pystate_to_npstate(pystate)
            rng = np_rng = np.random.RandomState(seed=0)
            np_rng.set_state(npstate)
    elif impl == 'python':
        if rng is None:
            rng = random
        elif isinstance(rng, int):
            rng = random.Random(rng)
        elif isinstance(rng, np.random.RandomState):
            # Convert numpy to python random state
            np_rng = rng
            npstate = np_rng.get_state()
            pystate = _npstate_to_pystate(npstate)
            rng = py_rng = random.Random(0)
            py_rng.setstate(pystate)
    else:
        raise KeyError('unknown rng impl={}'.format(impl))
    return rng


def random_indexes(max_index, subset_size=None, seed=None, rng=None):
    """ random unrepeated indicies

    Args:
        max_index (?):
        subset_size (None): (default = None)
        seed (None): (default = None)
        rng (RandomState):  random number generator(default = None)

    Returns:
        ?: subst

    CommandLine:
        python -m utool.util_numpy --exec-random_indexes

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_numpy import *  # NOQA
        >>> max_index = 10
        >>> subset_size = None
        >>> seed = None
        >>> rng = np.random.RandomState(0)
        >>> subst = random_indexes(max_index, subset_size, seed, rng)
        >>> result = ('subst = %s' % (str(subst),))
        >>> print(result)
    """
    subst_ = np.arange(0, max_index)
    rng = ensure_rng(seed if rng is None else rng)
    rng.shuffle(subst_)
    if subset_size is None:
        subst = subst_
    else:
        subst = subst_[0:min(subset_size, max_index)]
    return subst


#def list_index(search_list, to_find_list):
#    """ Keep this function
#    Searches search_list for each element in to_find_list"""
#    try:
#        toret = [np.where(search_list == item)[0][0] for item in to_find_list]
#    except IndexError as ex1:
#        print('ERROR: ' + str(ex1))
#        print('item = %r' % (item,))
#        raise
#    return toret


def npfind(arr):
    found = np.where(arr)[0]
    pos = -1 if len(found) == 0 else found[0]
    return pos


def index_of(item, array):
    'index of [item] in [array]'
    return np.where(array == item)[0][0]


def spaced_indexes(len_, n, trunc=False):
    """
    Returns n evenly spaced indexes.
    Returns as many as possible if trunc is true
    """

    if n is None:
        return np.arange(len_)
    all_indexes = np.arange(len_)
    if trunc:
        n = min(len_, n)
    if n == 0:
        return np.empty(0)
    stride = len_ // n
    try:
        indexes = all_indexes[0:-1:stride]
    except ValueError:
        raise ValueError('cannot slice list of len_=%r into n=%r parts' % (len_, n))
    return indexes


def inbounds(arr, low, high):
    flag_low = arr >= low
    flag_high = arr < high if high is not None else flag_low
    flag = np.logical_and(flag_low, flag_high)
    return flag


def intersect2d(A, B):
    """
    intersect2d

    intersect rows of 2d numpy arrays

    DEPRICATE: use intersect2d in vtool instead

    Args:
        A (ndarray[ndim=2]):
        B (ndarray[ndim=2]):

    Returns:
        tuple: (C, Ax, Bx)

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_numpy import *  # NOQA
        >>> import utool as ut
        >>> A = np.array([[1, 2, 3], [1, 1, 1]])
        >>> B = np.array([[1, 2, 3], [1, 2, 14]])
        >>> (C, Ax, Bx) = ut.intersect2d(A, B)
        >>> result = str((C, Ax, Bx))
        >>> print(result)
        (array([[1, 2, 3]]), array([0]), array([0]))
    """
    Cset  =  set(tuple(x) for x in A).intersection(set(tuple(x) for x in B))
    Ax = np.array([x for x, item in enumerate(A) if tuple(item) in Cset], dtype=np.int)
    Bx = np.array([x for x, item in enumerate(B) if tuple(item) in Cset], dtype=np.int)
    C = np.array(tuple(Cset))
    return C, Ax, Bx


#def unique_ordered(arr):
    #""" pandas.unique preseves order and seems to be faster due to index overhead """
    #import pandas as pd
    #return pd.unique(arr)
    #_, idx = np.unique(arr, return_index=True)
    #return arr[np.sort(idx)]


def deterministic_shuffle(list_, seed=0, rng=None):
    r"""
    Args:
        list_ (list):
        seed (int):

    Returns:
        list: list_

    CommandLine:
        python -m utool.util_numpy --test-deterministic_shuffle

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_numpy import *  # NOQA
        >>> list_ = [1, 2, 3, 4, 5, 6]
        >>> seed = 1
        >>> list_ = deterministic_shuffle(list_, seed)
        >>> result = str(list_)
        >>> print(result)
        [3, 2, 5, 1, 4, 6]
    """
    rng = ensure_rng(seed if rng is None else rng)
    rng.shuffle(list_)
    return list_


def shuffle(list_, seed=0, rng=None):
    r"""
    Shuffles a list inplace and then returns it for convinience

    Args:
        list_ (list or ndarray): list to shuffl
        rng (RandomState or int): seed or random number gen

    Returns:
        list: this is the input, but returned for convinience

    Example:
        >>> # ENABLE_DOCTEST
        >>> from utool.util_numpy import *  # NOQA
        >>> list1 = [1, 2, 3, 4, 5, 6]
        >>> list2 = shuffle(list(list1), rng=1)
        >>> assert list1 != list2
        >>> result = str(list2)
        >>> print(result)
        [3, 2, 5, 1, 4, 6]
    """
    rng = ensure_rng(seed if rng is None else rng)
    rng.shuffle(list_)
    return list_


def random_sample(list_, nSample, strict=False, rng=None, seed=None):
    """
    Grabs data randomly

    Args:
        list_ (list):
        nSample (?):
        strict (bool): (default = False)
        rng (module):  random number generator(default = numpy.random)
        seed (None): (default = None)

    Returns:
        list: sample_list

    CommandLine:
        python -m utool.util_numpy --exec-random_sample

    Example:
        >>> # DISABLE_DOCTEST
        >>> from utool.util_numpy import *  # NOQA
        >>> list_ = np.arange(10)
        >>> nSample = 4
        >>> strict = False
        >>> rng = np.random.RandomState(0)
        >>> seed = None
        >>> sample_list = random_sample(list_, nSample, strict, rng, seed)
        >>> result = ('sample_list = %s' % (str(sample_list),))
        >>> print(result)
    """
    rng = ensure_rng(seed if rng is None else rng)
    if isinstance(list_, list):
        list2_ = list_[:]
    else:
        list2_ = np.copy(list_)
    if len(list2_) == 0 and not strict:
        return list2_
    rng.shuffle(list2_)
    if nSample is None and strict is False:
        return list2_
    if not strict:
        nSample = min(max(0, nSample), len(list2_))
    sample_list = list2_[:nSample]
    return sample_list


def deterministic_sample(list_, nSample, seed=0, rng=None, strict=False):
    """ Grabs data randomly, but in a repeatable way """
    rng = ensure_rng(seed if rng is None else rng)
    sample_list = random_sample(list_, nSample, strict=strict, rng=rng)
    return sample_list


def spaced_items(list_, n, **kwargs):
    """ Returns n evenly spaced items """
    indexes = spaced_indexes(len(list_), n, **kwargs)
    items = list_[indexes]
    return items


def sample_domain(min_, max_, nSamp, mode='linear'):
    """
    Example:
        >>> # ENABLE_DOCTEST
        >>> import utool
        >>> min_ = 10
        >>> max_ = 1000
        >>> nSamp  = 7
        >>> result = utool.sample_domain(min_, max_, nSamp)
        >>> print(result)
        [10, 151, 293, 434, 576, 717, 859]
    """
    if mode == 'linear':
        samples_ = np.rint(np.linspace(min_, max_, nSamp + 1)).astype(np.int64)
    elif mode == 'log':
        base = 2
        logmin = np.log2(min_) / np.log2(base)
        logmax = np.log2(max_) / np.log2(base)
        samples_ = np.rint(np.logspace(logmin, logmax, nSamp + 1, base=base)).astype(np.int64)
    else:
        raise NotImplementedError(mode)
    sample = [index for index in samples_ if index < max_]
    return sample


def make_incrementer():
    # DEPRICATE FOR ITERTOOLS.COUNT
    return ut.partial(six.next, itertools.count(1))
    # def incrementer(_mem=[0]):
    #     _mem[0] += 1
    #     return _mem[0]
    # return incrementer


if __name__ == '__main__':
    """
    CommandLine:
        python -c "import utool, utool.util_numpy; utool.doctest_funcs(utool.util_numpy, allexamples=True)"
        python -c "import utool, utool.util_numpy; utool.doctest_funcs(utool.util_numpy)"
        python -m utool.util_numpy
        python -m utool.util_numpy --allexamples
    """
    import multiprocessing
    multiprocessing.freeze_support()  # for win32
    import utool as ut  # NOQA
    ut.doctest_funcs()



# -*- coding: utf-8 -*-
"""
TODO: in the future maybe this is a contribution to pip
"""
from __future__ import print_function, division, absolute_import, unicode_literals
import utool as ut


# def get_site_package_directories():
#     import site
#     import sys
#     import six
#     sitepackages = site.getsitepackages()
#     if sys.platform.startswith('darwin'):
#         if six.PY2:
#             macports_site = '/opt/local/Library/Frameworks/Python.framework/Versions/2.7/lib/python2.7/site-packages'
#         else:
#             macports_site = '/opt/local/Library/Frameworks/Python.framework/Versions/3.5/lib/python3.5/site-packages'
#             assert six.PY2, 'fix this for python 3'
#         sitepackages = [macports_site] + sitepackages
#     return sitepackages


def module_stdinfo_dict(module, versionattr='__version__', version=None,
                        libdep=None, name=None, **kwargs):
    infodict = ut.odict()
    if module is None:
        infodict['__version__'] = version
        infodict['__name__'] = name
        infodict['__file__'] = 'None'
    else:
        if version is not None:
            infodict['__version__'] = version
        else:
            infodict['__version__'] = getattr(module, versionattr, None)
        infodict['__name__'] = name
        infodict['__name__'] = getattr(module, '__name__', None)
        infodict['__file__'] = getattr(module, '__file__', None)

    if libdep is not None:
        infodict['libdep'] = libdep
    infodict.update(kwargs)
    return infodict


def print_module_info(modname):
    print('Checking modname = %r' % (modname,))
    # Handles special cases for certain modules
    if modname.lower() == 'pillow':
        from PIL import Image
        import PIL
        pil_path = PIL.__path__
        infodict = module_stdinfo_dict(Image, versionattr='PILLOW_VERSION',
                                       image_version=Image.VERSION,
                                       pil_path=pil_path)
    elif modname.lower() == 'pyqt4':
        from PyQt4 import QtCore
        infodict = module_stdinfo_dict(QtCore, 'PYQT_VERSION_STR')
    elif modname.lower() == 'pyqt5':
        from PyQt5 import QtCore
        infodict = module_stdinfo_dict(QtCore, 'PYQT_VERSION_STR')
    else:
        # Handle normal modules
        module = ut.import_modname(modname)
        infodict = module_stdinfo_dict(module)
    if any([infodict['__file__'].endswith(ext) for ext in ut.LIB_EXT_LIST]):
        infodict['libdep'] = ut.get_dynlib_dependencies(infodict['__file__'])
    return print(ut.repr4(infodict, strvals=True))


if __name__ == '__main__':
    r"""
    CommandLine:
        export PYTHONPATH=$PYTHONPATH:/home/joncrall/code/utool/utool/util_scripts
        python ~/code/utool/utool/util_scripts/pipinfo.py
        python ~/code/utool/utool/util_scripts/pipinfo.py --allexamples

        pipinfo.py networkx
        pipinfo.py sklearn
        pipinfo.py cv2
        pipinfo.py PyQt4
    """
    modname = ut.get_argval('--modname')
    if not modname:
        args = ut.get_cmdline_varargs()
        if len(args) > 0:
            modname = args[0]
    assert modname
    print_module_info(modname)
